import io
import os
import urllib.request
from datetime import datetime, timedelta
from dateutil import parser
import pytest
import pytz

from boxsdk import BoxAPIException
from test.integration_new.context_managers.box_retention_policy_assigment import BoxRetentionPolicyAssignment
from test.integration_new.context_managers.box_retention_policy import BoxRetentionPolicy
from test.integration_new import util
from test.integration_new.context_managers.box_test_file import BoxTestFile
from test.integration_new.context_managers.box_test_folder import BoxTestFolder


FILE_TESTS_DIRECTORY_NAME = 'file-integration-tests'


@pytest.fixture(scope="module", autouse=True)
def parent_folder():
    with BoxTestFolder(name=f'{FILE_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder:
        yield folder


@pytest.fixture(scope="module", autouse=True)
def test_file(parent_folder, small_file_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        yield file


def test_preflight_check(test_file):
    file_size = 1213
    accelerator_url = test_file.preflight_check(size=file_size, name=util.random_name())

    assert accelerator_url


def test_create_upload_session_for_file_version(test_file):
    file_size = 26000000
    upload_session = test_file.create_upload_session(file_size=file_size)

    assert upload_session.id
    assert upload_session.type == 'upload_session'


def test_get_chuncked_uploader_for_file_version(test_file, large_file):
    total_size = os.stat(large_file.path).st_size

    chunked_uploader = test_file.get_chunked_uploader(file_path=large_file.path)

    assert chunked_uploader._file_size == total_size
    assert chunked_uploader._upload_session


def test_content(test_file, small_file_path):
    file_content = test_file.content()

    with open(small_file_path, 'rb') as expected_file:
        assert expected_file.read() == file_content


def test_download_to(test_file, small_file_path):
    output_file_name = util.random_name()

    with open(output_file_name, 'wb') as output_file:
        test_file.download_to(writeable_stream=output_file)

    with open(output_file_name, 'rb') as downloaded_file:
        with open(small_file_path, 'rb') as expected_file:
            assert downloaded_file.read() == expected_file.read()

    os.remove(output_file_name)


def test_get_download_url(test_file, small_file_path):
    download_url = test_file.get_download_url()

    with open(small_file_path, 'rb') as expected_file:
        with urllib.request.urlopen(download_url) as downloaded_file:
            assert downloaded_file.read() == expected_file.read()


def test_update_contents_with_stream(parent_folder, small_file_path, small_file_v2_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        with open(small_file_v2_path, 'rb') as file_v2:
            updated_file = file.update_contents_with_stream(file_stream=file_v2)

        with open(small_file_v2_path, 'rb') as file_v2:
            assert updated_file.content() == file_v2.read()


def test_update_contents(parent_folder, small_file_path, small_file_v2_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        updated_file = file.update_contents(small_file_v2_path)

        with open(small_file_v2_path, 'rb') as file_v2:
            assert updated_file.content() == file_v2.read()


def test_lock_and_unlock(parent_folder, small_file_path, small_file_v2_path, other_user, other_client):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        file.collaborate(accessible_by=other_user, role='editor')

        try:
            file.lock(prevent_download=True)

            assert file.get(fields=('lock',)).lock

            with pytest.raises(BoxAPIException):
                other_client.file(file.object_id).update_contents(small_file_v2_path)

            with pytest.raises(BoxAPIException):
                other_client.file(file.object_id).download_to(io.BytesIO())

        finally:
            file.unlock()

        assert file.get(fields=('lock',)).lock is None


def test_get_shared_link(parent_folder, small_file_path, other_user, other_client):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        file.collaborate(accessible_by=other_user, role='editor')

        shared_link = other_client.file(file.object_id).get_shared_link(allow_edit=True, allow_preview=True, allow_download=True)

        result_permissions = file.get().shared_link['permissions']
        assert result_permissions == {'can_preview': True, 'can_download': True, 'can_edit': True}
        assert other_client.get_shared_item(shared_link).id == file.id


def test_get_shared_link_download_url(parent_folder, small_file_path, other_user, other_client):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        file.collaborate(accessible_by=other_user, role='editor')

        shared_link_download_url = other_client.file(file.object_id).get_shared_link_download_url()

        with open(small_file_path, 'rb') as expected_file:
            with urllib.request.urlopen(shared_link_download_url) as downloaded_file:
                assert downloaded_file.read() == expected_file.read()


def test_add_and_get_comments(parent_folder, small_file_path):
    test_comment = 'this is a test comment'
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        file.add_comment(message=test_comment)
        comment = list(file.get_comments())[0]

        assert comment.id
        assert comment.message == test_comment


def test_and_get_task(parent_folder, small_file_path):
    test_task_action = 'review'
    test_task_message = 'this is a test task message'
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        file.create_task(message=test_task_message, action=test_task_action)
        task = list(file.get_tasks())[0]

        assert task.id
        assert task.message == test_task_message
        assert task.action == test_task_action


def test_get_previous_versions(parent_folder, small_file_path, small_file_v2_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as test_file:
        test_file_version = test_file.file_version

        assert not list(test_file.get_previous_versions())

        updated_test_file = test_file.update_contents(small_file_v2_path)
        updated_file_version = updated_test_file.file_version

        previous_version = list(updated_test_file.get_previous_versions())[0]

        assert previous_version != updated_file_version
        assert previous_version == test_file_version
        assert previous_version.type == 'file_version'


def test_promote_version(parent_folder, small_file_path, small_file_v2_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as test_file:
        test_file_version = test_file.file_version

        updated_test_file = test_file.update_contents(small_file_v2_path)
        updated_file_version = updated_test_file.file_version

        assert updated_file_version != test_file_version

        promoted_file_version = updated_test_file.promote_version(file_version=test_file_version)

        assert test_file.get(fields=('file_version',)).file_version == promoted_file_version
        assert set(updated_test_file.get_previous_versions()) == {test_file_version, updated_file_version}

        with open(small_file_path, 'rb') as expected_file_v1:
            assert expected_file_v1.read() == test_file.content()


def test_delete_version(parent_folder, small_file_path, small_file_v2_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as test_file:
        updated_test_file = test_file.update_contents(small_file_v2_path)

        previous_version = list(updated_test_file.get_previous_versions())[0]
        assert previous_version.trashed_at is None

        was_deletion_successs = updated_test_file.delete_version(file_version=previous_version)

        assert was_deletion_successs
        assert list(test_file.get_previous_versions())[0].trashed_at


def test_get_embed_url(parent_folder, image_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=image_path) as test_file:
        embed_url = test_file.get_embed_url()

        assert embed_url


def test_get_representation_info(parent_folder, image_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=image_path) as test_file:
        representation_info = test_file.get_representation_info(rep_hints='[png]')

        assert representation_info


def test_get_thumbnail(parent_folder, image_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=image_path) as test_file:
        thumbnail = test_file.get_thumbnail(extension='jpg', min_width=32, max_width=32, min_height=32, max_height=32)

        assert thumbnail


def test_get_thumbnail_representation(parent_folder, image_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=image_path) as test_file:
        thumbnail_representation = test_file.get_thumbnail_representation(extension='jpg', dimensions='32x32')

        assert thumbnail_representation


def test_copy(test_file, parent_folder):
    copied_file_name = util.random_name()
    copied_file = test_file.copy(parent_folder=parent_folder, name=copied_file_name)

    try:
        assert copied_file.id != test_file
        assert copied_file.name == copied_file_name
        assert test_file.content() == copied_file.content()
    finally:
        util.permanently_delete(copied_file)


def test_set_disposition_at(parent_folder, small_file_path):
    with BoxRetentionPolicy(disposition_action='permanently_delete', retention_length=1) as retention_policy:
        with BoxTestFolder(name=f'{FILE_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder_under_retention:
            with BoxRetentionPolicyAssignment(retention_policy=retention_policy, assignee=folder_under_retention):
                with BoxTestFile(parent_folder=folder_under_retention, file_path=small_file_path) as file_under_retention:
                    old_disposition_str = file_under_retention.get(fields=('disposition_at',)).disposition_at
                    old_disposition_datetime = parser.parse(old_disposition_str)

                    new_disposition_date = datetime.now().replace(microsecond=0).astimezone(pytz.utc) + timedelta(days=2)
                    file_under_retention.set_disposition_at(new_disposition_date)

                    updated_disposition_str = file_under_retention.get(fields=('disposition_at',)).disposition_at
                    updated_disposition_datetime = parser.parse(updated_disposition_str)

                    assert updated_disposition_datetime.astimezone(pytz.utc) == new_disposition_date
                    assert updated_disposition_datetime.astimezone(pytz.utc) != old_disposition_datetime.astimezone(pytz.utc)



import json
from datetime import datetime
from io import BytesIO
from os.path import basename
from unittest.mock import mock_open, patch, Mock, MagicMock
import pytest
import pytz

from boxsdk.config import API
from boxsdk.exception import BoxAPIException
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.object.enterprise import Enterprise
from boxsdk.object.file import File
from boxsdk.object.metadata_cascade_policy import MetadataCascadePolicy
from boxsdk.object.web_link import WebLink
from boxsdk.object.collaboration import Collaboration, CollaborationRole
from boxsdk.object.folder import Folder, FolderSyncState
from boxsdk.object.upload_session import UploadSession
from boxsdk.session.box_response import BoxResponse
from boxsdk.util.chunked_uploader import ChunkedUploader


# pylint:disable=protected-access
# pylint:disable=redefined-outer-name
from boxsdk.util.datetime_formatter import normalize_date_to_rfc3339_format
from boxsdk.util.default_arg_value import SDK_VALUE_NOT_SET


@pytest.fixture()
def mock_new_upload_accelerator_url():
    return 'https://upload.box.com/api/2.0/files/content?upload_session_id=123'


@pytest.fixture(scope='function')
def mock_accelerator_response_for_new_uploads(make_mock_box_request, mock_new_upload_accelerator_url):
    mock_response, _ = make_mock_box_request(
        response={
            'upload_url': mock_new_upload_accelerator_url,
            'upload_token': None,
        }
    )
    return mock_response


@pytest.fixture()
def mock_items(mock_box_session, mock_object_id):
    return [
        {'type': 'file', 'id': mock_object_id},
        {'type': 'folder', 'id': mock_object_id},
        {'type': 'file', 'id': mock_object_id},
    ], [
        File(mock_box_session, mock_object_id),
        Folder(mock_box_session, mock_object_id),
        File(mock_box_session, mock_object_id),
    ]


@pytest.fixture()
def mock_items_response(mock_items):
    # pylint:disable=redefined-outer-name
    def get_response(limit, offset):
        items_json, items = mock_items
        entries = items_json[offset:limit + offset]
        mock_box_response = Mock(BoxResponse)
        mock_network_response = Mock(DefaultNetworkResponse)
        mock_box_response.network_response = mock_network_response
        mock_box_response.json.return_value = mock_json = {
            'entries': entries,
            'total_count': len(entries),
            'limit': limit,
            'offset': offset,
        }
        mock_box_response.content = json.dumps(mock_json).encode()
        mock_box_response.status_code = 200
        mock_box_response.ok = True
        return mock_box_response, items[offset:limit + offset]
    return get_response


def test_get_chunked_uploader(mock_box_session, mock_content_response, mock_file_path, test_folder):
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions'
    mock_file_stream = BytesIO(mock_content_response.content)
    file_size = 197520
    file_name = 'file'
    part_size = 12345
    total_parts = 16
    num_parts_processed = 0
    upload_session_type = 'upload_session'
    upload_session_id = 'F971964745A5CD0C001BBE4E58196BFD'
    expected_data = {
        'folder_id': test_folder.object_id,
        'file_size': file_size,
        'file_name': file_name,
    }
    mock_box_session.post.return_value.json.return_value = {
        'id': upload_session_id,
        'type': upload_session_type,
        'num_parts_processed': num_parts_processed,
        'total_parts': total_parts,
        'part_size': part_size,
    }
    with patch('os.stat') as stat:
        stat.return_value.st_size = file_size
        with patch('boxsdk.object.folder.open', return_value=mock_file_stream):
            chunked_uploader = test_folder.get_chunked_uploader(mock_file_path)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    upload_session = chunked_uploader._upload_session
    assert upload_session.part_size == part_size
    assert upload_session.total_parts == total_parts
    assert upload_session.num_parts_processed == num_parts_processed
    assert upload_session.type == upload_session_type
    assert upload_session.id == upload_session_id
    assert isinstance(chunked_uploader, ChunkedUploader)


@pytest.fixture()
def mock_items_response_with_marker(mock_items):
    # pylint:disable=redefined-outer-name
    def get_response(limit, offset):
        items_json, items = mock_items
        entries = items_json[offset:limit + offset]
        mock_box_response = Mock(BoxResponse)
        mock_network_response = Mock(DefaultNetworkResponse)
        mock_box_response.network_response = mock_network_response
        mock_box_response.json.return_value = mock_json = {
            'entries': entries,
            'total_count': len(entries),
            'limit': limit,
            'offset': offset,
        }
        mock_box_response.content = json.dumps(mock_json).encode()
        mock_box_response.status_code = 200
        mock_box_response.ok = True
        return mock_box_response, items[offset:limit + offset]
    return get_response


def _assert_collaborator_added(test_folder, collaborator, mock_box_session, mock_collab_response, notify, role, can_view_path, data):
    mock_box_session.post.return_value = mock_collab_response
    collaboration = test_folder.add_collaborator(collaborator, role, notify, can_view_path)
    assert isinstance(collaboration, Collaboration)
    expected_url = API.BASE_API_URL + '/collaborations'
    params = {'notify': notify}
    mock_box_session.post.assert_called_once_with(expected_url, expect_json_response=True, data=data, params=params)


@pytest.mark.parametrize('accessible_by', ['user', 'group', 'email'])
@pytest.mark.parametrize('notify', [True, False])
@pytest.mark.parametrize('role', iter(CollaborationRole))
@pytest.mark.parametrize('can_view_path', [True, False])
def test_add_collaborator(test_folder, mock_user, mock_group, mock_box_session, mock_collab_response, accessible_by, notify, role, can_view_path):
    accessible_dict = {
        'user': (mock_user, {'id': mock_user.object_id, 'type': 'user'}),
        'group': (mock_group, {'id': mock_group.object_id, 'type': 'group'}),
        'email': ('foo@example.com', {'login': 'foo@example.com', 'type': 'user'}),
    }

    invitee, mock_accessible_by = accessible_dict[accessible_by]

    body_params = {
        'item': {'id': test_folder.object_id, 'type': 'folder'},
        'accessible_by': mock_accessible_by,
        'role': role,
    }
    if can_view_path:
        body_params['can_view_path'] = True
    data = json.dumps(body_params)
    _assert_collaborator_added(test_folder, invitee, mock_box_session, mock_collab_response, notify, role, can_view_path, data)


def test_add_collaborator_raises_for_bad_type(test_folder):
    with pytest.raises(TypeError):
        test_folder.add_collaborator(b'byte string', CollaborationRole.EDITOR)


@pytest.mark.parametrize('recursive', [True, False])
def test_delete_folder(test_folder, mock_box_session, recursive, etag, if_match_header):
    test_folder.delete(recursive=recursive, etag=etag)
    expected_url = test_folder.get_url()
    mock_box_session.delete.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        params={'recursive': recursive},
        headers=if_match_header,
    )


@pytest.mark.parametrize('limit,offset,fields,sort,direction', [
    (1, 0, None, None, None),
    (100, 0, ['foo', 'bar'], None, None),
    (1, 1, None, None, None),
    (1, 0, None, 'name', 'ASC'),
    (1, 1, None, 'date', 'DESC')
])
def test_get_items(test_folder, mock_box_session, mock_items_response, limit, offset, fields, sort, direction):
    # pylint:disable=redefined-outer-name
    expected_url = test_folder.get_url('items')
    mock_box_session.get.return_value, expected_items = mock_items_response(limit, offset)
    items = test_folder.get_items(limit, offset, fields=fields, sort=sort, direction=direction)
    expected_params = {'limit': limit, 'offset': offset}
    if fields:
        expected_params['fields'] = ','.join(fields)
    if sort:
        expected_params['sort'] = sort
    if direction:
        expected_params['direction'] = direction
    for actual, expected in zip(items, expected_items):
        assert actual == expected
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params)
    assert all(i.id == e.object_id for (i, e) in zip(items, expected_items))


@pytest.mark.parametrize('content_created_at', [
    '1970-01-01T00:00:00+00:00',
    datetime(1970, 1, 1, 0, 0, 0, tzinfo=pytz.UTC),
    None
])
@pytest.mark.parametrize('content_modified_at', [
    '1970-01-01T11:11:11+00:00',
    datetime(1970, 1, 1, 11, 11, 11, tzinfo=pytz.UTC),
    None
])
@pytest.mark.parametrize('is_stream', (True, False))
def test_upload(
        test_folder,
        mock_box_session,
        mock_content_response,
        mock_upload_response,
        mock_file_path,
        mock_object_id,
        upload_using_accelerator,
        mock_accelerator_response_for_new_uploads,
        mock_new_upload_accelerator_url,
        upload_using_accelerator_fails,
        is_stream,
        etag,
        sha1,
        if_match_sha1_header,
        content_created_at,
        content_modified_at
):
    # pylint:disable=too-many-locals
    # pylint:disable=too-many-arguments
    file_description = 'Test File Description'
    additional_attributes = {'attr': 123}
    expected_url = f'{API.UPLOAD_URL}/files/content'
    if upload_using_accelerator:
        if upload_using_accelerator_fails:
            mock_box_session.options.side_effect = BoxAPIException(400)
        else:
            mock_box_session.options.return_value = mock_accelerator_response_for_new_uploads
            expected_url = mock_new_upload_accelerator_url

    mock_box_session.post.return_value = mock_upload_response

    if is_stream:
        mock_file_stream = BytesIO(mock_content_response.content)
        new_file = test_folder.upload_stream(
            mock_file_stream,
            basename(mock_file_path),
            file_description,
            upload_using_accelerator=upload_using_accelerator,
            content_created_at=content_created_at,
            content_modified_at=content_modified_at,
            additional_attributes=additional_attributes,
            sha1=sha1,
            etag=etag,
        )
    else:
        mock_file = mock_open(read_data=mock_content_response.content)
        mock_file_stream = mock_file.return_value
        with patch('boxsdk.object.folder.open', mock_file, create=True):
            new_file = test_folder.upload(
                mock_file_path,
                file_description=file_description,
                upload_using_accelerator=upload_using_accelerator,
                content_created_at=content_created_at,
                content_modified_at=content_modified_at,
                additional_attributes=additional_attributes,
                sha1=sha1,
                etag=etag,
            )

    mock_files = {'file': ('unused', mock_file_stream)}
    attributes = {
        'name': basename(mock_file_path),
        'parent': {'id': mock_object_id},
        'description': file_description,
        'content_created_at': normalize_date_to_rfc3339_format(content_created_at),
        'content_modified_at': normalize_date_to_rfc3339_format(content_modified_at),
    }
    # Using `update` to mirror the actual impl, since the attributes could otherwise come through in a different order
    # in Python 2 tests
    attributes.update(additional_attributes)
    data = {'attributes': json.dumps(attributes)}
    mock_box_session.post.assert_called_once_with(expected_url, expect_json_response=False, files=mock_files, data=data, headers=if_match_sha1_header)
    assert isinstance(new_file, File)
    assert new_file.object_id == mock_object_id
    assert 'id' in new_file
    assert new_file['id'] == mock_object_id
    assert new_file.description == file_description
    assert not hasattr(new_file, 'entries')
    assert 'entries' not in new_file


@pytest.mark.parametrize('is_stream', (True, False))
def test_upload_combines_preflight_and_accelerator_calls_if_both_are_requested(
        test_folder,
        mock_box_session,
        mock_file_path,
        mock_content_response,
        mock_accelerator_response_for_new_uploads,
        is_stream
):
    mock_box_session.options.return_value = mock_accelerator_response_for_new_uploads

    if is_stream:
        mock_file_stream = BytesIO(mock_content_response.content)
        test_folder.upload_stream(
            mock_file_stream,
            basename(mock_file_path),
            preflight_check=True,
            upload_using_accelerator=True,
        )
    else:
        mock_file = mock_open(read_data=mock_content_response.content)
        with patch('boxsdk.object.folder.open', mock_file, create=True):
            test_folder.upload(
                mock_file_path,
                preflight_check=True,
                upload_using_accelerator=True,
            )

    mock_box_session.options.assert_called_once()


def test_create_upload_session(test_folder, mock_box_session):
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions'
    file_size = 197520
    file_name = 'test_file.pdf'
    upload_session_id = 'F971964745A5CD0C001BBE4E58196BFD'
    upload_session_type = 'upload_session'
    num_parts_processed = 0
    total_parts = 16
    part_size = 12345
    expected_data = {
        'folder_id': test_folder.object_id,
        'file_size': file_size,
        'file_name': file_name,
    }
    mock_box_session.post.return_value.json.return_value = {
        'id': upload_session_id,
        'type': upload_session_type,
        'num_parts_processed': num_parts_processed,
        'total_parts': total_parts,
        'part_size': part_size,
    }
    upload_session = test_folder.create_upload_session(file_size, file_name)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert isinstance(upload_session, UploadSession)
    assert upload_session.part_size == part_size
    assert upload_session.total_parts == total_parts
    assert upload_session.num_parts_processed == num_parts_processed
    assert upload_session.type == upload_session_type
    assert upload_session.id == upload_session_id


def test_upload_stream_does_preflight_check_if_specified(
        mock_box_session,
        test_folder,
        preflight_check,
        preflight_fails,
        file_size,
):
    with patch.object(Folder, 'preflight_check', return_value=None):
        kwargs = {'file_stream': BytesIO(b'some bytes'), 'file_name': 'foo.txt'}
        mock_box_session.post = MagicMock()
        if preflight_check:
            kwargs['preflight_check'] = preflight_check
            kwargs['preflight_expected_size'] = file_size
        if preflight_fails:
            test_folder.preflight_check.side_effect = BoxAPIException(400)
            with pytest.raises(BoxAPIException):
                test_folder.upload_stream(**kwargs)
        else:
            test_folder.upload_stream(**kwargs)

        if preflight_check:
            assert test_folder.preflight_check.called_once_with(size=file_size, name='foo.txt')
            _assert_post_called_correctly(mock_box_session, preflight_fails)
        else:
            assert not test_folder.preflight_check.called


def _assert_post_called_correctly(mock_box_session, preflight_fails):
    if preflight_fails:
        assert not mock_box_session.post.called
    else:
        assert mock_box_session.post.called


@patch('boxsdk.object.folder.open', mock_open(read_data=b'some bytes'), create=True)
def test_upload_does_preflight_check_if_specified(
        mock_box_session,
        test_folder,
        mock_file_path,
        preflight_check,
        preflight_fails,
        file_size,
):
    with patch.object(Folder, 'preflight_check', return_value=None):
        kwargs = {'file_path': mock_file_path, 'file_name': 'foo.txt'}
        mock_box_session.post = MagicMock()
        if preflight_check:
            kwargs['preflight_check'] = preflight_check
            kwargs['preflight_expected_size'] = file_size
        if preflight_fails:
            test_folder.preflight_check.side_effect = BoxAPIException(400)
            with pytest.raises(BoxAPIException):
                test_folder.upload(**kwargs)
        else:
            test_folder.upload(**kwargs)

        if preflight_check:
            assert test_folder.preflight_check.called_once_with(size=file_size, name='foo.txt')
            _assert_post_called_correctly(mock_box_session, preflight_fails)
        else:
            assert not test_folder.preflight_check.called


def test_create_subfolder(test_folder, mock_box_session, mock_object_id, mock_folder_response):
    expected_url = test_folder.get_type_url()
    mock_box_session.post.return_value = mock_folder_response
    new_folder = test_folder.create_subfolder('name')
    data = json.dumps({'name': 'name', 'parent': {'id': mock_object_id}})
    mock_box_session.post.assert_called_once_with(expected_url, data=data)
    assert isinstance(new_folder, Folder)
    assert new_folder.object_id == mock_object_id


def test_get_shared_link(
        test_folder,
        mock_box_session,
        shared_link_access,
        shared_link_unshared_at,
        shared_link_password,
        shared_link_can_download,
        shared_link_can_preview,
        shared_link_vanity_name,
        test_url,
        etag,
        if_match_header,
):
    # pylint:disable=redefined-outer-name, protected-access
    expected_url = test_folder.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_folder.object_type,
        'id': test_folder.object_id,
        'shared_link': {
            'url': test_url,
        },
    }
    expected_data = {'shared_link': {}}
    if shared_link_access is not None:
        expected_data['shared_link']['access'] = shared_link_access
    if shared_link_unshared_at is not SDK_VALUE_NOT_SET:
        expected_data['shared_link']['unshared_at'] = normalize_date_to_rfc3339_format(shared_link_unshared_at)
    if shared_link_can_download is not None or shared_link_can_preview is not None:
        expected_data['shared_link']['permissions'] = permissions = {}
        if shared_link_can_download is not None:
            permissions['can_download'] = shared_link_can_download
        if shared_link_can_preview is not None:
            permissions['can_preview'] = shared_link_can_preview
    if shared_link_password is not None:
        expected_data['shared_link']['password'] = shared_link_password
    if shared_link_vanity_name is not None:
        expected_data['shared_link']['vanity_name'] = shared_link_vanity_name

    url = test_folder.get_shared_link(
        etag=etag,
        access=shared_link_access,
        unshared_at=shared_link_unshared_at,
        password=shared_link_password,
        allow_download=shared_link_can_download,
        allow_preview=shared_link_can_preview,
        vanity_name=shared_link_vanity_name,
    )
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_data),
        headers=if_match_header,
        params=None,
    )
    assert url == test_url


@pytest.mark.parametrize('sync_state', iter(FolderSyncState))
def test_update_sync_state(test_folder, mock_folder_response, mock_box_session, sync_state):
    expected_url = test_folder.get_url()
    mock_box_session.put.return_value = mock_folder_response
    data = {'sync_state': sync_state}
    update_response = test_folder.update_sync_state(sync_state)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), params=None, headers=None)
    assert isinstance(update_response, Folder)
    assert update_response.object_id == test_folder.object_id


def test_preflight(
        test_folder,
        mock_object_id,
        mock_box_session,
        mock_accelerator_response_for_new_uploads,
        mock_new_upload_accelerator_url,
):
    new_file_size, new_file_name = 100, 'foo.txt'
    mock_box_session.options.return_value = mock_accelerator_response_for_new_uploads

    accelerator_url = test_folder.preflight_check(size=new_file_size, name=new_file_name)

    mock_box_session.options.assert_called_once_with(
        url=f'{API.BASE_API_URL}/files/content',
        expect_json_response=True,
        data=json.dumps(
            {
                'size': new_file_size,
                'name': new_file_name,
                'parent': {'id': mock_object_id},
            }
        ),
    )
    assert accelerator_url == mock_new_upload_accelerator_url


def test_create_web_link_returns_the_correct_web_link_object(test_folder, mock_box_session):
    expected_url = f"{API.BASE_API_URL}/web_links"
    expected_name = 'Test WebLink'
    description = 'Test Description'
    test_web_link_url = 'https://test.com'
    mock_box_session.post.return_value.json.return_value = {
        'type': 'web_link',
        'id': '42',
        'url': test_web_link_url,
        'name': expected_name,
        'description': description
    }
    new_web_link = test_folder.create_web_link(test_web_link_url, expected_name, description)
    data = {
        'url': test_web_link_url,
        'parent': {
            'id': '42',
        },
        'name': expected_name,
        'description': description,
    }
    mock_box_session.post.assert_called_once_with(
        expected_url,
        data=json.dumps(data),
    )
    assert isinstance(new_web_link, WebLink)
    assert new_web_link.object_id == '42'
    assert new_web_link.url == test_web_link_url
    assert new_web_link.name == expected_name
    assert new_web_link.description == description


def test_get_metadata_cascade_policies(test_folder, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_cascade_policies'
    params = {'folder_id': test_folder.object_id}
    mock_box_session.get.return_value.json.return_value = {
        'entries': [
            {
                'id': '84113349-794d-445c-b93c-d8481b223434',
                'type': 'metadata_cascade_policy',
                'parent': {
                    'type': 'folder',
                    'id': test_folder.object_id,
                },
                'scope': 'enterprise_11111',
                'templateKey': 'testTemplate',
            }
        ],
        'next_marker': None,
        'prev_marker': None,
    }

    cascade_policies = test_folder.get_metadata_cascade_policies()
    policy = cascade_policies.next()

    mock_box_session.get.assert_called_once_with(expected_url, params=params)
    assert isinstance(policy, MetadataCascadePolicy)
    assert policy.object_id == '84113349-794d-445c-b93c-d8481b223434'
    assert policy.scope == 'enterprise_11111'
    assert policy.templateKey == 'testTemplate'
    # pylint: disable=protected-access
    assert policy._session == mock_box_session


def test_cascade_metadata(test_folder, mock_box_session, test_metadata_template):
    expected_url = f'{API.BASE_API_URL}/metadata_cascade_policies'
    expected_body = {
        'folder_id': test_folder.object_id,
        'scope': test_metadata_template.scope,
        'templateKey': test_metadata_template.template_key,
    }
    mock_box_session.post.return_value.json.return_value = {
        'id': '84113349-794d-445c-b93c-d8481b223434',
        'type': 'metadata_cascade_policy',
        'owner_enterprise': {
            'type': 'enterprise',
            'id': '11111',
        },
        'parent': {
            'type': 'folder',
            'id': test_folder.object_id,
        },
        'scope': test_metadata_template.scope,
        'templateKey': test_metadata_template.template_key,
    }

    cascade_policy = test_folder.cascade_metadata(test_metadata_template)

    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(cascade_policy, MetadataCascadePolicy)
    assert cascade_policy.object_id == '84113349-794d-445c-b93c-d8481b223434'
    enterprise = cascade_policy.owner_enterprise
    assert isinstance(enterprise, Enterprise)
    assert enterprise.object_id == '11111'
    folder = cascade_policy.parent
    assert isinstance(folder, Folder)
    assert folder.object_id == test_folder.object_id
    assert cascade_policy.scope == test_metadata_template.scope
    assert cascade_policy.templateKey == test_metadata_template.template_key


def test_get_folder_locks(test_folder, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/folder_locks'
    params = {'folder_id': test_folder.object_id}
    mock_box_session.get.return_value.json.return_value = {
        "entries": [
            {
                "folder": {
                    "id": "12345",
                    "etag": "1",
                    "type": "folder",
                    "sequence_id": "3",
                    "name": "Contracts"
                },
                "id": "12345678",
                "type": "folder_lock",
                "created_by": {
                    "id": "11446498",
                    "type": "user"
                },
                "created_at": "2020-09-14T23:12:53Z",
                "locked_operations": {
                    "move": True,
                    "delete": True
                },
                "lock_type": "freeze"
            }
        ],
        "limit": 1000,
        "next_marker": None
    }

    folder_locks = test_folder.get_locks()
    lock = folder_locks.next()

    mock_box_session.get.assert_called_once_with(expected_url, params=params)
    assert lock.id == '12345678'
    assert lock.folder.id == '12345'
    assert lock.locked_operations['move']
    # pylint: disable=protected-access
    assert lock._session == mock_box_session


def test_create_folder_lock(test_folder, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/folder_locks'
    expected_body = {
        "folder": {
            "type": "folder",
            "id": test_folder.object_id
        },
        "locked_operations": {
            "move": True,
            "delete": True
        }
    }
    mock_box_session.post.return_value.json.return_value = {
        "id": "12345678",
        "type": "folder_lock",
        "created_at": "2020-09-14T23:12:53Z",
        "created_by": {
            "id": "11446498",
            "type": "user"
        },
        "folder": {
            "id": "12345",
            "type": "folder",
            "etag": "1",
            "name": "Contracts",
            "sequence_id": "3"
        },
        "lock_type": "freeze",
        "locked_operations": {
            "delete": True,
            "move": True
        }
    }

    lock = test_folder.create_lock()

    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert lock.id == '12345678'
    assert lock.folder.id == '12345'
    assert lock.locked_operations['move']
    # pylint: disable=protected-access
    assert lock._session == mock_box_session


def test_delete_folder_lock(test_folder_lock, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/folder_locks/{test_folder_lock.object_id}'
    test_folder_lock.delete()
    mock_box_session.delete.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        headers=None,
        params={}
    )



from typing import Any, Iterable

from boxsdk.object.metadata_template import MetadataTemplate, MetadataField, MetadataFieldType
from test.integration_new import CLIENT


class BoxTestMetadataTemplate:

    def __init__(self, *, display_name: str, fields: Iterable = None):
        if fields is None:
            fields = []
        self._metadata_template: MetadataTemplate = CLIENT.create_metadata_template(display_name=display_name, fields=fields)

    def __enter__(self) -> MetadataTemplate:
        return self._metadata_template

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._metadata_template.delete()



import json
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from boxsdk.util.datetime_formatter import normalize_date_to_rfc3339_format
from boxsdk.util.text_enum import TextEnum

from ..util.api_call_decorator import api_call
from .base_object import BaseObject

if TYPE_CHECKING:
    from boxsdk.object.folder import Folder


class StatusState(TextEnum):
    """An enum of possible status states"""
    ACTIVE = 'active'
    INACTIVE = 'inactive'


class FileRequest(BaseObject):
    """Represents the file request."""
    _item_type = 'file_request'

    @api_call
    def copy(
            self,
            *,
            folder: 'Folder',
            description: Optional[str] = None,
            expires_at: Union[datetime, str] = None,
            require_description: Optional[bool] = None,
            require_email: Optional[bool] = None,
            status: Optional[str] = None,
            title: Optional[str] = None,
            **_kwargs
    ) -> 'FileRequest':
        """Copy an existing file request already present on one folder, and applies it to another folder.

        :param description:
            A new description for the file request.
        :param title:
            A new title for the file request.
        :param expires_at:
            A expiration time for file request which no longer accepts new files.
        :param folder:
            The folder to which the file request will be saved.
        :param require_description:
            A flag indicating whether the file submitted must have a description.
        :param require_email:
            A flag indicating whether the file submitted must have sender email.
        :param status:
            The status of the file request.
        :returns:
            The copy of the file request
        """
        url = self.get_url('copy')
        data = {
            'folder': {'id': folder.object_id, 'type': folder.object_type},
        }
        if description is not None:
            data['description'] = description
        if title is not None:
            data['title'] = title
        if expires_at is not None:
            data['expires_at'] = normalize_date_to_rfc3339_format(expires_at)
        if require_description is not None:
            data['is_description_required'] = require_description
        if require_email is not None:
            data['is_email_required'] = require_email
        if status is not None:
            data['status'] = status
        box_response = self._session.post(url, data=json.dumps(data))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )



from datetime import datetime
from os.path import dirname, join

from bottle import Bottle, debug, request, response, view, TEMPLATE_PATH
import sqlalchemy
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import StaticPool

from test.functional.mock_box.behavior.event_behavior import EventBehavior
from test.functional.mock_box.behavior.file_behavior import FileBehavior
from test.functional.mock_box.behavior.folder_behavior import FolderBehavior
from test.functional.mock_box.behavior.oauth2_behavior import OAuth2Behavior
from test.functional.mock_box.behavior.user_behavior import UserBehavior
from test.functional.mock_box.db_model import DbModel
# pylint:disable=unused-import
from test.functional.mock_box.db_model.collaboration_model import CollaborationModel
from test.functional.mock_box.db_model.event_model import EventModel
from test.functional.mock_box.db_model.file_model import FileModel
from test.functional.mock_box.db_model.folder_model import FolderModel
from test.functional.mock_box.db_model.group_model import GroupModel
from test.functional.mock_box.db_model.lock_model import LockModel
from test.functional.mock_box.db_model.share_model import ShareFileModel, ShareFolderModel
# pylint:enable=unused-import
from test.functional.mock_box.db_model.user_model import UserModel
from test.functional.mock_box.util.chaos_utils import allow_chaos
from test.functional.mock_box.util.http_utils import (
    authorize,
    log_request,
    rate_limit,
    retry_after,
    DELETE,
    GET,
    OPTIONS,
    POST,
    PUT,
    StoppableWSGIRefServer,
)


TEMPLATE_PATH.insert(0, join(dirname(__file__), 'views'))


class Box:
    """
    Fake Box. Sets up 4 webservers - one for auth, one for upload, one for events, and one for the rest of the api.
    """
    API_PORT = 18086
    UPLOAD_PORT = 18087
    OAUTH_API_PORT = 18088
    EVENT_PORT = 18089
    OAUTH_AUTHORIZE_PORT = 18090
    RATE_LIMIT_THRESHOLD = 100
    RATE_LIMIT_REQUEST_PER_SECOND = 4

    def __init__(self):
        debug(True)
        self._db_engine = None
        self._db_session = None
        self._db_session_maker = None
        self.reset_filesystem()
        # Mock Box consists of 3 webservers - one for the content API, one for the upload API, and one for OAuth2
        api, upload, oauth_api, event, oauth_authorize = Bottle(), Bottle(), Bottle(), Bottle(), Bottle()
        app_mapping = {
            self.API_PORT: api,
            self.EVENT_PORT: event,
            self.OAUTH_API_PORT: oauth_api,
            self.UPLOAD_PORT: upload,
            self.OAUTH_AUTHORIZE_PORT: oauth_authorize,
        }
        # Since we don't instantiate the servers until Box is instantiated, we have to apply the routes now
        for routed_method in (getattr(self, m) for m in dir(self) if hasattr(getattr(self, m), 'route')):
            app_port = routed_method.app
            app = app_mapping[app_port]
            app.route(routed_method.route, routed_method.verb, routed_method)
        for code in [400, 401, 404, 409, 429, 500]:
            for app in app_mapping.values():
                app.error(code)(self.handle_error)
        self._api = StoppableWSGIRefServer(host='localhost', port=self.API_PORT).run(api)
        self._upload = StoppableWSGIRefServer(host='localhost', port=self.UPLOAD_PORT).run(upload)
        self._oauth_api = StoppableWSGIRefServer(host='localhost', port=self.OAUTH_API_PORT).run(oauth_api)
        self._event = StoppableWSGIRefServer(host='localhost', port=self.EVENT_PORT).run(event)
        self._oauth_authorize = StoppableWSGIRefServer(host='localhost', port=self.OAUTH_AUTHORIZE_PORT).run(oauth_authorize)
        self._rate_limit_bucket = (self.RATE_LIMIT_THRESHOLD, datetime.utcnow())

    @staticmethod
    def handle_error(error):
        response.content_type = 'application/json'
        return error.body

    def shutdown(self):
        """Shutdown the webservers and wait for them to exit."""
        self._api.shutdown()
        self._upload.shutdown()
        self._oauth_api.shutdown()
        self._oauth_authorize.shutdown()
        self._api.wait()
        self._upload.wait()
        self._oauth_api.wait()
        self._oauth_authorize.wait()

    def reset_filesystem(self, users=(), applications=()):
        """
        Create in-memory DB that can be accessed by multiple threads.
        Set up auth requests, the rate limit bucket, and the request log.
        """
        self._db_engine = sqlalchemy.create_engine(
            'sqlite:///:memory:',
            connect_args={'check_same_thread': False},
            poolclass=StaticPool,
        )
        DbModel.metadata.create_all(self._db_engine)
        self._db_session_maker = sessionmaker(bind=self._db_engine, autoflush=True)
        self._db_session = self._db_session_maker()
        self._db_session.add(FolderModel(folder_id=0))
        self._db_session.commit()
        self._rate_limit_bucket = (self.RATE_LIMIT_THRESHOLD, datetime.utcnow())
        self._request_log = []
        self._oauth_behavior = OAuth2Behavior(self._db_session)
        self._file_behavior = FileBehavior(self._db_session)
        self._folder_behavior = FolderBehavior(self._db_session)
        self._event_behavior = EventBehavior(self._db_session)
        self._user_behavior = UserBehavior(self._db_session)
        user_ids = []
        for user_info in users:
            user_name, user_login = user_info
            user_id = self.add_user(user_name, user_login)
            user_ids.append(user_id)
        for app_info in applications:
            client_id, client_secret, user_index = app_info
            self.add_application(client_id, client_secret, user_ids[user_index])

    @property
    def requests(self):
        return self._request_log

    @property
    def oauth(self):
        return self._oauth_behavior

    def add_application(self, client_id, client_secret, user_ids):
        return self._oauth_behavior.add_application(client_id, client_secret, user_ids)

    def add_user(self, name, login):
        user = UserModel(name=name, login=login)
        self._db_session.add(user)
        self._db_session.commit()
        return user.user_id

    def check_authorization_header(self):
        """
        Check that the request has an auth header and that its token matches the currently valid token.
        Further check that the token isn't expired.

        Called by methods decorated with the authorize decorator.
        """
        return self._oauth_behavior.check_authorization_header()

    def check_rate_limits(self):
        """
        Implements the token bucket algorithm, whereby incoming requests remove tokens from a fixed capacity bucket
        that is refilling at the steady state rate limit.
        In this case, the bucket capacity is RATE_LIMIT_THRESHOLD and the refill rate is RATE_LIMIT_REQUEST_PER_SECOND.

        Requests that are over the rate limit are aborted with 429 (Too Many Requests). A Retry-After header is
        specified that estimates when another request will succeed.
        """
        capactiy, timestamp = self._rate_limit_bucket
        now = datetime.utcnow()
        delta = self.RATE_LIMIT_REQUEST_PER_SECOND * (now - timestamp).microseconds / 1e6
        new_capacity = min(self.RATE_LIMIT_THRESHOLD, capactiy + delta)
        self._rate_limit_bucket = (new_capacity - 1, now)
        if new_capacity < 1:
            delay = (-(new_capacity - 4) / self.RATE_LIMIT_REQUEST_PER_SECOND) * 2
            retry_after(delay)

    def append_to_request_log(self):
        self._request_log.append((request.method, request.route))

    @log_request
    @allow_chaos
    @GET(OAUTH_AUTHORIZE_PORT, '/')
    @view('oauth2')
    def oauth2_authorize(self):
        return self._oauth_behavior.oauth2_authorize()

    @log_request
    @allow_chaos
    @POST(OAUTH_AUTHORIZE_PORT, '/')
    def oauth2_finish_loop(self):
        return self._oauth_behavior.oauth2_finish_loop()

    @log_request
    @allow_chaos
    @POST(OAUTH_API_PORT, '/token')
    def oauth2_token(self):
        return self._oauth_behavior.oauth2_token()

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @POST(UPLOAD_PORT, '/files/<file_id>/content')
    def update_file(self, file_id):
        return self._file_behavior.update_file(file_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @POST(UPLOAD_PORT, '/files/content')
    def upload_file(self):
        return self._file_behavior.upload_file()

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @GET(API_PORT, '/files/<file_id>')
    def get_file_info(self, file_id):
        return self._file_behavior.get_file_info(file_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @PUT(API_PORT, '/files/<file_id>')
    def update_file_info(self, file_id):
        return self._file_behavior.update_file_info(file_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @POST(API_PORT, '/files/<file_id>/copy')
    def copy_file(self, file_id):
        return self._file_behavior.copy_file(file_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @GET(API_PORT, '/files/<file_id>/content')
    def download_file(self, file_id):
        return self._file_behavior.download_file(file_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @DELETE(API_PORT, '/files/<file_id>')
    def delete_file(self, file_id):
        return self._file_behavior.delete_file(file_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @GET(API_PORT, '/folders/<folder_id>')
    def get_folder_info(self, folder_id):
        return self._folder_behavior.get_folder_info(folder_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @PUT(API_PORT, '/folders/<folder_id>')
    def update_folder_info(self, folder_id):
        return self._folder_behavior.update_folder_info(folder_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @POST(API_PORT, '/folders/<folder_id>/copy')
    def copy_folder(self, folder_id):
        return self._folder_behavior.copy_folder(folder_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @DELETE(API_PORT, '/folders/<folder_id>')
    def delete_folder(self, folder_id):
        return self._folder_behavior.delete_folder(folder_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @GET(API_PORT, '/folders/<folder_id>/items')
    def get_folder_items(self, folder_id):
        return self._folder_behavior.get_folder_items(folder_id)

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @POST(API_PORT, '/folders')
    def create_folder(self):
        return self._folder_behavior.create_folder()

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @GET(API_PORT, '/events')
    def get_events(self):
        return self._event_behavior.get_events()

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @OPTIONS(API_PORT, '/events')
    def get_long_poll_url(self):
        return self._event_behavior.get_long_poll_url()

    @log_request
    @allow_chaos
    @GET(EVENT_PORT, '/subscribe')
    def subscribe(self):
        return self._event_behavior.subscribe()

    @log_request
    @allow_chaos
    @GET(EVENT_PORT, '/check')
    def check(self):
        return self._event_behavior.subscribe()

    @log_request
    @rate_limit
    @authorize
    @allow_chaos
    @GET(API_PORT, '/users/<user_id>')
    def get_user_info(self, user_id):
        return self._user_behavior.get_user_info(user_id)



import hashlib
from concurrent.futures import ThreadPoolExecutor, as_completed, Future
from threading import Lock
from typing import IO, TYPE_CHECKING, Optional, List

from boxsdk.exception import BoxException
from boxsdk.config import API

if TYPE_CHECKING:
    from boxsdk.object.file import File
    from boxsdk.object.upload_session import UploadSession


class ChunkedUploader:

    def __init__(self, upload_session: 'UploadSession', content_stream: IO[bytes], file_size: int):
        """
        The initializer for the :class:`ChunkedUploader`

        :param upload_session:
            The upload session for doing the chunked uploader.
        :param content_stream:
            The file-like object to upload.
        :param file_size:
            The total size of the file for the chunked upload.
        :returns:
            An intialized`ChunkedUploader` object.
        """
        self._upload_session = upload_session
        self._content_stream = content_stream
        self._file_size = file_size
        self._part_array = []
        self._sha1 = hashlib.sha1()
        self._part_definitions = {}
        self._is_aborted = False
        self._lock = Lock()
        self._inflight_parts = {}
        self._chunk_index = 0
        self._executor = ThreadPoolExecutor(max_workers=API.CHUNK_UPLOAD_THREADS)

    def start(self) -> Optional['File']:
        """
        Starts the process of chunk uploading a file. Should return file. If commit was not processed will return None.
        You can call ChunkedUploader.resume to retry committing upload.

        :returns:
            An uploaded :class:`File` or None if session was not processed
        """
        if self._is_aborted:
            raise BoxException('The upload has been previously aborted. Please retry upload with a new upload session.')

        futures = [self._executor.submit(self._upload_part) for _ in range(self._upload_session.total_parts)]
        self._upload(futures)
        return self._commit_and_erase_stream_reference_when_succeed()

    def resume(self) -> Optional['File']:
        """
        Resumes the process of chunk uploading a file from where upload failed.
        Should return file. If commit was not processed will return None.
        You can call ChunkedUploader.resume to retry committing upload.

        :returns:
            An uploaded :class:`File` or None if session was not processed
        """
        if self._is_aborted:
            raise BoxException('The upload has been previously aborted. Please retry upload with a new upload session.')

        self._executor = ThreadPoolExecutor(max_workers=API.CHUNK_UPLOAD_THREADS)
        parts = self._upload_session.get_parts()
        for part in parts:
            self._part_definitions[part['offset']] = part

        with self._lock:
            futures = [self._executor.submit(lambda item=part: self._upload_part(item)) for part in self._inflight_parts.values()]
            for _ in range(self._upload_session.total_parts - self._chunk_index - len(self._inflight_parts)):
                futures.append(self._executor.submit(self._upload_part))

        self._upload(futures)
        return self._commit_and_erase_stream_reference_when_succeed()

    def abort(self) -> bool:
        """
        Abort an upload session, cancelling the upload and removing any parts that have already been uploaded.

        :returns:
            A boolean indication success of the upload abort.
        """
        self._content_stream = None
        self._part_definitions = {}
        self._inflight_parts = {}
        self._is_aborted = True
        return self._upload_session.abort()

    def _upload(self, futures: List[Future]) -> None:
        """
        Utility function for looping through all parts of the upload session and uploading them.
        """
        try:
            for future in as_completed(futures):
                future.result()
        except Exception as exc:
            self._executor.shutdown(wait=True)
            raise exc

        self._part_array = sorted(self._part_definitions.values(), key=lambda part: part['offset'])

    def _upload_part(self, task=None):
        if isinstance(task, InflightPart):
            next_part = task
        else:
            with self._lock:
                next_part = self._get_next_part()
                self._sha1.update(next_part.chunk)

        with self._lock:
            if self._part_definitions.get(next_part.offset):
                return
            self._inflight_parts[next_part.offset] = next_part

        uploaded_part = next_part.upload()

        with self._lock:
            self._part_definitions[next_part.offset] = uploaded_part
            del self._inflight_parts[next_part.offset]

    def _get_next_part(self) -> 'InflightPart':
        """
        Retrieves the next :class:`InflightPart` that needs to be uploaded

        :returns:
            The :class:`InflightPart` object to be uploaded next.
        """
        copied_length = 0
        chunk = b''
        offset = self._chunk_index * self._upload_session.part_size
        self._chunk_index += 1
        while copied_length < self._upload_session.part_size:
            bytes_read = self._content_stream.read(self._upload_session.part_size - copied_length)
            if bytes_read is None:
                # stream returns none when no bytes are ready currently but there are
                # potentially more bytes in the stream to be read.
                continue
            if not bytes_read:
                # stream is exhausted.
                break
            chunk += bytes_read
            copied_length += len(bytes_read)
        return InflightPart(offset, chunk, self._upload_session, self._file_size)

    def _commit_and_erase_stream_reference_when_succeed(self):
        content_sha1 = self._sha1.digest()
        commit_result = self._upload_session.commit(content_sha1=content_sha1, parts=self._part_array)
        # Remove file stream reference when uploading file succeeded
        if commit_result is not None:
            self._content_stream = None
        self._executor.shutdown(wait=True)
        return commit_result


class InflightPart:

    def __init__(self, offset: int, chunk: bytes, upload_session: 'UploadSession', total_size: int):
        """
        The initializer for the :class:`InflightPart` object.

        :param offset:
            The offset for the :class:`InflightPart` that represents the position of the part to be uploaded
        :param chunk:
            The chunk in bytes to be uploaded.
        :param upload_session:
            The :class:`UploadSession` for the :class:`InflightPart`.
        :param total_size:
            The total size of the file to be chunked uploaded.
        """
        self._offset = offset
        self._chunk = chunk
        self._upload_session = upload_session
        self._total_size = total_size

    @property
    def offset(self) -> int:
        """
        Getter for the offset of the :class:`InflightPart`
        """
        return self._offset

    @property
    def chunk(self) -> bytes:
        """
        Getter for the chunk of the :class:`InflightPart`
        """
        return self._chunk

    def upload(self) -> dict:
        """
        Upload method for the :class:`InflightPart`

        :returns:
            The uploaded part record.
        """
        return self._upload_session.upload_part_bytes(
            part_bytes=self.chunk,
            offset=self.offset,
            total_size=self._total_size
        )



from boxsdk.object.legal_hold import LegalHold
from boxsdk.config import API


def test_get(test_legal_hold, mock_box_session):
    file_version_id = '1234'
    expected_url = f'{API.BASE_API_URL}/file_version_legal_holds/{test_legal_hold.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'legal_hold',
        'id': test_legal_hold.object_id,
        'file_version': {
            'type': 'file_version',
            'id': file_version_id,
        },
    }
    file_version_legal_hold = test_legal_hold.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(file_version_legal_hold, LegalHold)
    assert file_version_legal_hold['file_version']['id'] == file_version_id
    assert file_version_legal_hold.type == 'legal_hold'
    assert file_version_legal_hold.id == test_legal_hold.object_id



from unittest.mock import Mock
import json
import pytest

from boxsdk.object.legal_hold_policy import LegalHoldPolicy
from boxsdk.object.legal_hold_policy_assignment import LegalHoldPolicyAssignment
from boxsdk.config import API
from boxsdk.network.default_network import DefaultNetworkResponse


@pytest.fixture(scope='module')
def policy_id_1():
    return 101


@pytest.fixture(scope='module')
def policy_id_2():
    return 202


@pytest.fixture(scope='module')
def legal_hold_id_1():
    return 101


@pytest.fixture(scope='module')
def legal_hold_id_2():
    return 202


@pytest.fixture(scope='module')
def policies_response(policy_id_1, policy_id_2):
    # pylint disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'legal_hold_policy', 'id': policy_id_1, 'name': 'Test Policy 1'},
            {'type': 'legal_hold_policy', 'id': policy_id_2, 'name': 'Test Policy 2'}
        ],
        'limit': 5,
    }
    return mock_network_response


@pytest.fixture(scope='module')
def legal_hold_response(legal_hold_id_1, legal_hold_id_2):
    # pylint disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'legal_hold', 'id': legal_hold_id_1, 'name': 'Test Legal Hold 1'},
            {'type': 'legal_hold', 'id': legal_hold_id_2, 'name': 'Test Legal Hold 2'}
        ],
        'limit': 5,
    }
    return mock_network_response


def test_assign(test_legal_hold_policy, mock_box_session, test_file):
    assignment_id = '12345'
    assigned_at = '2016-05-18T17:38:03-07:00'
    expected_url = f'{API.BASE_API_URL}/legal_hold_policy_assignments'
    expected_body = {
        'policy_id': test_legal_hold_policy.object_id,
        'assign_to': {
            'type': 'file',
            'id': test_file.object_id
        }
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': 'legal_hold_policy_assignment',
        'id': assignment_id,
        'assigned_at': assigned_at,
    }
    new_legal_hold_assignment = test_legal_hold_policy.assign(test_file)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(new_legal_hold_assignment, LegalHoldPolicyAssignment)
    assert new_legal_hold_assignment.assigned_at == assigned_at
    assert new_legal_hold_assignment.id == assignment_id


def test_get(test_legal_hold_policy, mock_box_session):
    created_at = '2016-05-18T17:38:03-07:00'
    expected_url = f'{API.BASE_API_URL}/legal_hold_policies/{test_legal_hold_policy.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'legal_hold_policy',
        'id': test_legal_hold_policy.object_id,
        'created_at': created_at
    }
    legal_hold_policy = test_legal_hold_policy.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(legal_hold_policy, LegalHoldPolicy)
    assert legal_hold_policy.created_at == created_at


@pytest.fixture(scope='module')
def delete_policy_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_delete_policy_return_the_correct_response(
        test_legal_hold_policy,
        mock_box_session,
        delete_policy_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value = delete_policy_response
    response = test_legal_hold_policy.delete()
    # pylint:disable=protected-access
    expected_url = test_legal_hold_policy.get_url()
    # pylint:enable = protected-access
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)
    assert response is True


@pytest.mark.parametrize('assign_to_type, assign_to_id, params', [
    (None, None, {}),
    ('file', None, {'assign_to_type': 'file'}),
    ('folder', '22222', {'assign_to_type': 'folder', 'assign_to_id': '22222'})
])
def test_get_assignments(
        test_legal_hold_policy,
        mock_box_session,
        policies_response,
        policy_id_1,
        policy_id_2,
        assign_to_type,
        assign_to_id,
        params,
):
    # pylint:disable=redefined-outer-name
    expected_url = f'{API.BASE_API_URL}/legal_hold_policy_assignments'
    expected_params = {'policy_id': test_legal_hold_policy.object_id}
    expected_params.update(params)
    mock_box_session.get.return_value = policies_response
    assignments = test_legal_hold_policy.get_assignments(assign_to_type=assign_to_type, assign_to_id=assign_to_id)
    for assignment, expected_id in zip(assignments, [policy_id_1, policy_id_2]):
        assert assignment.object_id == expected_id
        # pylint:disable=protected-access
        assert assignment._session == mock_box_session
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params)


def test_get_file_version_legal_holds(
        test_legal_hold_policy,
        mock_box_session,
        legal_hold_response,
        legal_hold_id_1,
        legal_hold_id_2
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value = legal_hold_response
    legal_holds = test_legal_hold_policy.get_file_version_legal_holds()
    for legal_hold, expected_id in zip(legal_holds, [legal_hold_id_1, legal_hold_id_2]):
        assert legal_hold.object_id == expected_id
        # pylint:disable=protected-access
        assert legal_hold._session == mock_box_session



import json

from boxsdk.object.base_object import BaseObject
from ..util.api_call_decorator import api_call


class Comment(BaseObject):
    """An object that represents a comment on an item"""
    _item_type = 'comment'

    @staticmethod
    def construct_params_from_message(message: str) -> dict:
        message_type = 'tagged_message' if '@[' in message else 'message'
        return {
            message_type: message
        }

    @api_call
    def reply(self, message: str) -> 'Comment':
        """
        Add a reply to the comment.

        :param message:
            The content of the reply comment.
        """
        url = self.get_type_url()
        data = self.construct_params_from_message(message)
        data['item'] = {
            'type': 'comment',
            'id': self.object_id
        }
        box_response = self._session.post(url, data=json.dumps(data))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def edit(self, message: str) -> 'Comment':
        """
        Edit the message of the comment.

        :param message:
            The content of the reply comment.
        """
        data = self.construct_params_from_message(message)
        return self.update_info(data=data)



import pytest

from boxsdk.util.multipart_stream import MultipartStream


@pytest.fixture(params=({}, {'data_1': b'data_1_value', 'data_2': b'data_2_value'}))
def multipart_stream_data(request):
    return request.param


@pytest.fixture(params=({}, {'file_1': b'file_1_value', 'file_2': b'file_2_value'}))
def multipart_stream_files(request):
    return request.param


def test_multipart_stream_orders_data_before_files(multipart_stream_data, multipart_stream_files):
    # pylint:disable=redefined-outer-name
    if not multipart_stream_data and not multipart_stream_files:
        pytest.xfail('Encoder does not support empty fields.')
    stream = MultipartStream(multipart_stream_data, multipart_stream_files)
    encoded_stream = stream.to_string()
    data_indices = [encoded_stream.find(value) for value in multipart_stream_data.values()]
    file_indices = [encoded_stream.find(value) for value in multipart_stream_files.values()]
    assert -1 not in data_indices
    assert -1 not in file_indices
    assert all(all(data_index < f for f in file_indices) for data_index in data_indices)



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from boxsdk import NetworkResponse


class BoxResponse:
    """Represents a response to a Box API request."""

    def __init__(self, network_response):
        self._network_response = network_response

    def json(self) -> Any:
        """
        Return the parsed JSON response.
        """
        return self._network_response.json()

    @property
    def content(self) -> Any:
        """
        Return the content of the response body.
        """
        return self._network_response.content

    @property
    def ok(self) -> bool:
        """
        Return whether or not the request was successful.
        """
        # pylint:disable=invalid-name
        return self._network_response.ok

    @property
    def status_code(self) -> int:
        """
        Return the HTTP status code of the response.
        """
        return self._network_response.status_code

    @property
    def headers(self) -> dict:
        """
        Get the response headers.
        """
        return self._network_response.headers

    @property
    def network_response(self) -> 'NetworkResponse':
        """
        Return the underlying network response.
        """
        return self._network_response

    def __repr__(self) -> str:
        return f'<Box Response[{self.status_code}]>'



from bottle import request
from test.functional.mock_box.util.db_utils import get_folder_by_id
from test.functional.mock_box.util.http_utils import abort
from test.functional.mock_box.util import json_utils as json


class ItemBehavior:
    def __init__(self, db_session):
        self._db_session = db_session

    def _get_parent(self):
        params = json.load(request.body)
        parent = params.get('parent')
        if parent is None or 'id' not in parent:
            abort(400, 'Missing parameter: parent(id)')
        parent_id = parent['id']
        return get_folder_by_id(self._db_session, parent_id)



import os.path
import uuid
import warnings
from pathlib import Path

from test.integration_new import CLIENT
from boxsdk.exception import BoxAPIException
from boxsdk.object.base_item import BaseItem


def permanently_delete(item: BaseItem):
    item.delete(recursive=True)
    try:
        CLIENT.trash().permanently_delete_item(item)
    except BoxAPIException:
        warnings.warn(
            f"Unable to permanently remove {item.type}: {item.id} from trash. Probably this item is under retention.",
            category=RuntimeWarning,
            stacklevel=2
        )


def get_current_dir_path() -> str:
    return str(Path(os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))))


def get_file_path(file_name: str) -> str:
    return str(Path(f"{os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))}/resources/{file_name}"))


def random_name():
    return str(uuid.uuid4())



from test.integration_new.context_managers.box_test_user import BoxTestUser
from test.integration_new.context_managers.local_large_file import LocalLargeFile
from test.integration_new import util, CLIENT
import pytest


@pytest.fixture(scope='package')
def small_file_name():
    return 'small.pdf'


@pytest.fixture(scope='package')
def small_file_path(small_file_name):
    return util.get_file_path(small_file_name)


@pytest.fixture(scope='package')
def small_file_v2_name():
    return 'small_v2.pdf'


@pytest.fixture(scope='package')
def small_file_v2_path(small_file_v2_name):
    return util.get_file_path(small_file_v2_name)


@pytest.fixture(scope='package')
def image_name():
    return 'image.png'


@pytest.fixture(scope='package')
def image_path(image_name):
    return util.get_file_path(image_name)


@pytest.fixture(scope='session')
def large_file_name():
    return f'{util.random_name()}.pdf'


@pytest.fixture(scope="session")
def large_file(large_file_name):
    with LocalLargeFile(name=large_file_name) as large_file:
        yield large_file


@pytest.fixture(scope="module")
def user():
    with BoxTestUser(login=f'{util.random_name()}@box.com') as user:
        yield user


@pytest.fixture(scope="package")
def other_user():
    with BoxTestUser(login=None) as other_user:
        yield other_user


@pytest.fixture(scope="package")
def other_client(other_user):
    yield CLIENT.as_user(other_user)



from codecs import open   # pylint:disable=redefined-builtin
from os.path import dirname, join
import re
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: Implementation :: CPython',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Operating System :: OS Independent',
    'Operating System :: POSIX',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: MacOS :: MacOS X',
    'Topic :: Software Development :: Libraries :: Python Modules',
]


class PyTest(TestCommand):
    # pylint:disable=attribute-defined-outside-init

    user_options = [(b'pytest-args=', b'a', b"Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Do the import here, once the eggs are loaded.
        # pylint:disable=import-outside-toplevel
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def main():
    base_dir = dirname(__file__)
    install_requires = [
        'attrs>=17.3.0',
        'urllib3<2',
        'requests>=2.4.3,<3',
        'requests-toolbelt>=0.4.0',
        'python-dateutil',  # To be removed after dropping Python 3.6
    ]
    redis_requires = ['redis>=2.10.3']
    jwt_requires = ['pyjwt>=1.7.0', 'cryptography>=3']
    coveralls_requires = ['coveralls']
    dev_requires = ['tox<=3.28.0']
    gh_requires = ['tox-gh-actions']
    test_requires = [
        'bottle',
        'jsonpatch>1.14',
        'sqlalchemy<1.4.0',
        'pytest',
        'pytest-timeout',
        'pytest-cov',
        'pytest-lazy-fixture',
        'pytz',
    ]
    extra_requires = {
        'jwt': jwt_requires,
        'redis': redis_requires,
        'coveralls': coveralls_requires + dev_requires,
        'dev': dev_requires,
        'gh': gh_requires + dev_requires,
        'test': test_requires,
    }
    with open('boxsdk/version.py', encoding='utf-8') as config_py:
        version = re.search(r'^\s*__version__\s*=\s*[\'"]([^\'"]*)[\'"]', config_py.read(), re.MULTILINE).group(1)
    setup(
        name='boxsdk',
        version=version,
        description='Official Box Python SDK',
        long_description_content_type="text/markdown",
        long_description=open(join(base_dir, 'README.md'), encoding='utf-8').read(),  # pylint:disable=consider-using-with
        author='Box',
        author_email='oss@box.com',
        url='https://github.com/box/box-python-sdk',
        project_urls={
            'Changelog': 'https://github.com/box/box-python-sdk/blob/main/CHANGELOG.md',
        },
        packages=find_packages(exclude=['demo', 'docs', 'test', 'test*', '*test', '*test*']),
        install_requires=install_requires,
        extras_require=extra_requires,
        tests_require=test_requires,
        cmdclass={'test': PyTest},
        classifiers=CLASSIFIERS,
        keywords='box oauth2 sdk',
        license='Apache Software License, Version 2.0, http://www.apache.org/licenses/LICENSE-2.0',
        package_data={'boxsdk': ['py.typed']},
    )


if __name__ == '__main__':
    main()



import json

from boxsdk.config import API
from boxsdk.object.metadata_template import MetadataTemplate, MetadataField, MetadataFieldType


def test_get(test_metadata_template, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_templates/{test_metadata_template.scope}/{test_metadata_template.template_key}/schema'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'metadata_template',
        'scope': 'enterprise',
        'displayName': 'Vendor Contract',
        'hidden': True,
        'fields': [
            {
                'type': 'date',
                'displayName': 'Birthday',
                'key': 'bday',
            },
            {
                'type': 'enum',
                'displayName': 'State',
                'options': [
                    {'key': 'CA'},
                    {'key': 'TX'},
                    {'key': 'NY'},
                ],
            },
        ],
        'templateKey': 'vContract',
    }

    template = test_metadata_template.get()

    mock_box_session.get.assert_called_once_with(expected_url, params=None, headers=None)
    assert isinstance(template, MetadataTemplate)
    assert template.object_id is None
    assert template.displayName == 'Vendor Contract'
    fields = template.fields
    assert len(fields) == 2
    field = fields[0]
    assert isinstance(field, dict)
    assert field['type'] == 'date'
    assert field['key'] == 'bday'


def test_delete(test_metadata_template, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_templates/{test_metadata_template.scope}/{test_metadata_template.template_key}/schema'
    mock_box_session.delete.return_value.ok = True

    result = test_metadata_template.delete()

    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=None, params={})
    assert result is True


def test_update_info(test_metadata_template, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_templates/{test_metadata_template.scope}/{test_metadata_template.template_key}/schema'
    updates = test_metadata_template.start_update()
    updates.add_enum_option('state', 'WI')
    updates.add_field(
        MetadataField(field_type=MetadataFieldType.STRING, display_name='Name', description="Description", hidden=False)
    )
    updates.reorder_enum_options('state', ['CA', 'NY', 'TX', 'WI'])
    updates.reorder_fields(['bday', 'name', 'state'])
    updates.edit_field('state', MetadataField(None, 'State of Residency'))
    updates.edit_enum_option_key('state', 'WI', 'WY')
    updates.remove_enum_option('state', 'NY')
    updates.remove_field('bday')
    updates.edit_template({'hidden': False})

    expected_body = [
        {
            'op': 'addEnumOption',
            'fieldKey': 'state',
            'data': {'key': 'WI'},
        },
        {
            'op': 'addField',
            'data': {
                'type': 'string',
                'displayName': 'Name',
                "description": "Description",
                "hidden": False
            },
        },
        {
            'op': 'reorderEnumOptions',
            'fieldKey': 'state',
            'enumOptionKeys': ['CA', 'NY', 'TX', 'WI'],
        },
        {
            'op': 'reorderFields',
            'fieldKeys': ['bday', 'name', 'state'],
        },
        {
            'op': 'editField',
            'fieldKey': 'state',
            'data': {
                'displayName': 'State of Residency',
            },
        },
        {
            'op': 'editEnumOption',
            'fieldKey': 'state',
            'enumOptionKey': 'WI',
            'data': {
                'key': 'WY',
            },
        },
        {
            'op': 'removeEnumOption',
            'fieldKey': 'state',
            'enumOptionKey': 'NY',
        },
        {
            'op': 'removeField',
            'fieldKey': 'bday',
        },
        {
            'op': 'editTemplate',
            'data': {'hidden': False},
        },
    ]

    mock_box_session.put.return_value.json.return_value = {
        'type': 'metadata_template',
        'scope': 'enterprise',
        'displayName': 'Vendor Contract',
        'hidden': False,
        'fields': [
            {
                'type': 'string',
                'key': 'name',
                'displayName': 'Name',
                'description': 'Description',
                'hidden': False
            },
            {
                'type': 'enum',
                'key': 'state',
                'displayName': 'State of Residency',
                'options': [
                    {'key': 'CA'},
                    {'key': 'TX'},
                    {'key': 'WY'},
                ],
            },
        ],
        'templateKey': 'vContract',
    }

    updated_template = test_metadata_template.update_info(updates=updates)

    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(expected_body), headers=None, params=None)
    assert isinstance(updated_template, MetadataTemplate)
    assert updated_template.hidden is False
    assert updated_template.object_id is None
    fields = updated_template.fields
    assert len(fields) == 2
    field = fields[1]
    assert field['type'] == 'enum'
    assert field['displayName'] == 'State of Residency'
    assert len(field['options']) == 3
    assert field['options'][2]['key'] == 'WY'



import json
import os
from datetime import datetime
from typing import TYPE_CHECKING, Any, Tuple, Optional, Iterable, IO, Union

from boxsdk.object.group import Group
from boxsdk.object.item import Item
from boxsdk.object.user import User
from boxsdk.pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from boxsdk.pagination.marker_based_object_collection import MarkerBasedObjectCollection
from boxsdk.util.api_call_decorator import api_call
from boxsdk.util.datetime_formatter import normalize_date_to_rfc3339_format
from boxsdk.util.default_arg_value import SDK_VALUE_NOT_SET
from boxsdk.util.text_enum import TextEnum

if TYPE_CHECKING:
    from boxsdk.object.upload_session import UploadSession
    from boxsdk.util.chunked_uploader import ChunkedUploader
    from boxsdk.object.file import File
    from boxsdk.object.collaboration import CollaborationRole, Collaboration
    from boxsdk.object.web_link import WebLink
    from boxsdk.object.enterprise import Enterprise
    from boxsdk.pagination.box_object_collection import BoxObjectCollection
    from boxsdk.object.metadata_template import MetadataTemplate
    from boxsdk.object.metadata_cascade_policy import MetadataCascadePolicy
    from boxsdk.object.folder_lock import FolderLock


class FolderSyncState(TextEnum):
    """An enum of all possible values of a folder's ``sync_state`` attribute.

    The value of the ``sync_state`` attribute determines whether the folder
    will be synced by sync clients.
    """
    IS_SYNCED = 'synced'
    NOT_SYNCED = 'not_synced'
    PARTIALLY_SYNCED = 'partially_synced'


class _CollaborationType(TextEnum):
    """The type of a collaboration"""
    USER = 'user'
    GROUP = 'group'


class _Collaborator:
    """This helper class represents a collaborator on Box. A Collaborator can be a User, Group, or an email address"""

    def __init__(self, collaborator: Any):
        if isinstance(collaborator, User):
            self._setup(user=collaborator)
        elif isinstance(collaborator, Group):
            self._setup(group=collaborator)
        elif isinstance(collaborator, str):
            self._setup(email_address=collaborator)
        else:
            raise TypeError('Collaborator must be User, Group, or unicode string')

    def _setup(self, user: User = None, group: Group = None, email_address: str = None) -> None:
        """
        :param user:
            The Box user if applicable
        :param group:
            The Box group if applicable
        :param email_address:
            The email address of the user if not a user of Box
        """
        self._type = _CollaborationType.GROUP if group else _CollaborationType.USER
        id_object = user or group
        if id_object:
            self._key = 'id'
            self._identifier = id_object.object_id
        else:
            self._key = 'login'
            self._identifier = email_address

    @property
    def access(self) -> Tuple[str, str]:
        """Return a tuple for how to access collaborator

        The first element is the key for access, the second is the value
        """
        return self._key, self._identifier

    @property
    def type(self) -> str:
        """Return the type of collaborator (user or group)"""
        return self._type


class Folder(Item):
    """Box API endpoint for interacting with folders."""

    _item_type = 'folder'

    @api_call
    def preflight_check(self, size: int, name: str) -> Optional[str]:
        """
        Make an API call to check if a new file with given name and size can be uploaded to this folder.
        Returns an accelerator URL if one is available.

        :param size:
            The size of the file in bytes. Specify 0 for unknown file-sizes.
        :param name:
            The name of the file to be uploaded.
        :return:
            The Accelerator upload url or None if cannot get the Accelerator upload url.
        :raises:
            :class:`BoxAPIException` when preflight check fails.
        """
        return self._preflight_check(
            size=size,
            name=name,
            parent_id=self._object_id,
        )

    @api_call
    def create_upload_session(self, file_size: int, file_name: str) -> 'UploadSession':
        """
        Creates a new chunked upload session for upload a new file.

        :param file_size:
            The size of the file in bytes that will be uploaded.
        :param file_name:
            The name of the file that will be uploaded.
        :returns:
            A :class:`UploadSession` object.
        """
        url = f'{self.session.api_config.UPLOAD_URL}/files/upload_sessions'
        body_params = {
            'folder_id': self.object_id,
            'file_size': file_size,
            'file_name': file_name,
        }
        response = self._session.post(url, data=json.dumps(body_params)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_chunked_uploader(self, file_path: str, file_name: Optional[str] = None) -> 'ChunkedUploader':
        # pylint: disable=consider-using-with
        """
        Instantiate the chunked upload instance and create upload session with path to file.

        :param file_path:
            The local path to the file you wish to upload.
         :param file_name:
            The name with extention of the file that will be uploaded, e.g. new_file_name.zip.
            If not specified, the name from the local system is used.
        :returns:
            A :class:`ChunkedUploader` object.
        """
        total_size = os.stat(file_path).st_size
        upload_file_name = file_name if file_name else os.path.basename(file_path)
        content_stream = open(file_path, 'rb')

        try:
            upload_session = self.create_upload_session(total_size, upload_file_name)
            return upload_session.get_chunked_uploader_for_stream(content_stream, total_size)
        except Exception:
            content_stream.close()
            raise

    def _get_accelerator_upload_url_fow_new_uploads(self) -> Optional[str]:
        """
        Get Accelerator upload url for uploading new files.

        :return:
            The Accelerator upload url or None if cannot get one
        """
        return self._get_accelerator_upload_url()

    @api_call
    def get_items(
            self,
            limit: Optional[int] = None,
            offset: int = 0,
            marker: Optional[str] = None,
            use_marker: bool = False,
            sort: Optional[str] = None,
            direction: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> Iterable[Item]:
        """
        Get the items in a folder.

        :param limit:
            The maximum number of items to return per page. If not specified, then will use the server-side default.
        :param offset:
            The index at which to start returning items when using offset-based pagin.
        :param marker:
            The paging marker to start returning items from when using marker-based paging.
        :param use_marker:
            Whether to use marker-based paging instead of offset-based paging, defaults to False.
        :param sort:
            Item field to sort results on: 'id', 'name', or 'date'.
        :param direction:
            Sort direction for the items returned.
        :param fields:
            List of fields to request.
        :returns:
            The collection of items in the folder.
        """
        url = self.get_url('items')
        additional_params = {}
        if limit is not None:
            additional_params['limit'] = limit
        if sort:
            additional_params['sort'] = sort
        if direction:
            additional_params['direction'] = direction

        if use_marker:
            additional_params['usemarker'] = True
            return MarkerBasedObjectCollection(
                url=url,
                session=self._session,
                limit=limit,
                marker=marker,
                fields=fields,
                additional_params=additional_params,
                return_full_pages=False,
            )

        return LimitOffsetBasedObjectCollection(
            url=url,
            session=self._session,
            limit=limit,
            offset=offset,
            fields=fields,
            additional_params=additional_params,
            return_full_pages=False,
        )

    @api_call
    def upload_stream(
            self,
            file_stream: IO[bytes],
            file_name: str,
            file_description: Optional[str] = None,
            preflight_check: bool = False,
            preflight_expected_size: int = 0,
            upload_using_accelerator: bool = False,
            content_created_at: Union[datetime, str] = None,
            content_modified_at: Union[datetime, str] = None,
            additional_attributes: Optional[dict] = None,
            sha1: Optional[str] = None,
            etag: Optional[str] = None,
    ) -> 'File':
        """
        Upload a file to the folder.
        The contents are taken from the given file stream, and it will have the given name.

        :param file_stream:
            The file-like object containing the bytes
        :param file_name:
            The name to give the file on Box.
        :param file_description:
            The description to give the file on Box.
        :param preflight_check:
            If specified, preflight check will be performed before actually uploading the file.
        :param preflight_expected_size:
            The size of the file to be uploaded in bytes, which is used for preflight check. The default value is '0',
            which means the file size is unknown.
        :param upload_using_accelerator:
            If specified, the upload will try to use Box Accelerator to speed up the uploads for big files.
            It will make an extra API call before the actual upload to get the Accelerator upload url, and then make
            a POST request to that url instead of the default Box upload url. It falls back to normal upload endpoint,
            if cannot get the Accelerator upload url.

            Please notice that this is a premium feature, which might not be available to your app.
        :param content_created_at:
            A datetime string in a format supported by the dateutil library or a datetime.datetime object,
            which specifies when the file was created. If no timezone info provided, local timezone will be applied.
        :param content_modified_at:
            A datetime string in a format supported by the dateutil library or a datetime.datetime object, which
            specifies when the file was last modified. If no timezone info provided, local timezone will be applied.
        :param additional_attributes:
            A dictionary containing attributes to add to the file that are not covered by other parameters.
        :param sha1:
            A sha1 checksum for the file.
        :param etag:
            If specified, instruct the Box API to update the item only if the current version's etag matches.
        :returns:
            The newly uploaded file.
        """
        accelerator_upload_url = None
        if preflight_check:
            # Preflight check does double duty, returning the accelerator URL if one is available in the response.
            accelerator_upload_url = self.preflight_check(size=preflight_expected_size, name=file_name)
        elif upload_using_accelerator:
            accelerator_upload_url = self._get_accelerator_upload_url_fow_new_uploads()

        url = f'{self._session.api_config.UPLOAD_URL}/files/content'
        if upload_using_accelerator and accelerator_upload_url:
            url = accelerator_upload_url

        attributes = {
            'name': file_name,
            'parent': {'id': self._object_id},
            'description': file_description,
            'content_created_at': normalize_date_to_rfc3339_format(content_created_at),
            'content_modified_at': normalize_date_to_rfc3339_format(content_modified_at),
        }
        if additional_attributes:
            attributes.update(additional_attributes)

        data = {'attributes': json.dumps(attributes)}
        files = {
            'file': ('unused', file_stream),
        }
        headers = {}
        if etag is not None:
            headers['If-Match'] = etag
        if sha1 is not None:
            # The Content-MD5 field accepts sha1
            headers['Content-MD5'] = sha1
        if not headers:
            headers = None
        file_response = self._session.post(
            url, data=data, files=files, expect_json_response=False, headers=headers
        ).json()
        if 'entries' in file_response:
            file_response = file_response['entries'][0]
        return self.translator.translate(
            session=self._session,
            response_object=file_response,
        )

    @api_call
    def upload(
            self,
            file_path: str = None,
            file_name: str = None,
            file_description: Optional[str] = None,
            preflight_check: bool = False,
            preflight_expected_size: int = 0,
            upload_using_accelerator: bool = False,
            content_created_at: Union[datetime, str] = None,
            content_modified_at: Union[datetime, str] = None,
            additional_attributes: Optional[dict] = None,
            sha1: Optional[str] = None,
            etag: Optional[str] = None,
    ) -> 'File':
        """
        Upload a file to the folder.
        The contents are taken from the given file path, and it will have the given name.
        If file_name is not specified, the uploaded file will take its name from file_path.

        :param file_path:
            The file path of the file to upload to Box.
        :param file_name:
            The name to give the file on Box. If None, then use the leaf name of file_path
        :param file_description:
            The description to give the file on Box. If None, then no description will be set.
        :param preflight_check:
            If specified, preflight check will be performed before actually uploading the file.
        :param preflight_expected_size:
            The size of the file to be uploaded in bytes, which is used for preflight check. The default value is '0',
            which means the file size is unknown.
        :param upload_using_accelerator:
            If specified, the upload will try to use Box Accelerator to speed up the uploads for big files.
            It will make an extra API call before the actual upload to get the Accelerator upload url, and then make
            a POST request to that url instead of the default Box upload url. It falls back to normal upload endpoint,
            if cannot get the Accelerator upload url.

            Please notice that this is a premium feature, which might not be available to your app.
        :param content_created_at:
            A datetime string in a format supported by the dateutil library or a datetime.datetime object,
            which specifies when the file was created. If no timezone info provided, local timezone will be applied.
        :param content_modified_at:
            A datetime string in a format supported by the dateutil library or a datetime.datetime object, which
            specifies when the file was last modified.If no timezone info provided, local timezone will be applied.
        :param additional_attributes:
            A dictionary containing attributes to add to the file that are not covered by other parameters.
        :param sha1:
            A sha1 checksum for the new content.
        :param etag:
            If specified, instruct the Box API to update the item only if the current version's etag matches.
        :returns:
            The newly uploaded file.
        """
        if file_name is None:
            file_name = os.path.basename(file_path)
        with open(file_path, 'rb') as file_stream:
            return self.upload_stream(
                file_stream,
                file_name,
                file_description,
                preflight_check,
                preflight_expected_size=preflight_expected_size,
                upload_using_accelerator=upload_using_accelerator,
                content_created_at=content_created_at,
                content_modified_at=content_modified_at,
                additional_attributes=additional_attributes,
                sha1=sha1,
                etag=etag,
            )

    @api_call
    def create_subfolder(self, name: str) -> 'Folder':
        """
        Create a subfolder with the given name in the folder.

        :param name:
            The name of the new folder
        """
        url = self.get_type_url()
        data = {
            'name': name,
            'parent': {
                'id': self._object_id,
            }
        }
        box_response = self._session.post(url, data=json.dumps(data))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def update_sync_state(self, sync_state: FolderSyncState) -> 'Folder':
        """Update the ``sync_state`` attribute of this folder.

        Change whether this folder will be synced by sync clients.

        :param sync_state:
            The desired sync state of this folder.
            Must be a member of the `FolderSyncState` enum.
        :return:
            A new :class:`Folder` instance with updated information reflecting the new sync state.
        """
        data = {
            'sync_state': sync_state,
        }
        return self.update_info(data=data)

    @api_call
    def create_shared_link(
            self,
            *,
            access: Optional[str] = None,
            etag: Optional[str] = None,
            unshared_at: Union[datetime, str, None] = SDK_VALUE_NOT_SET,
            allow_download: Optional[bool] = None,
            allow_preview: Optional[bool] = None,
            password: Optional[str] = None,
            vanity_name: Optional[str] = None,
            **kwargs: Any
    ) -> 'Folder':
        """
        Baseclass override.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :param etag:
            If specified, instruct the Box API to create the link only if the current version's etag matches.
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
            The time portion can be omitted, which defaults to midnight (00:00:00) on that date.
        :param allow_download:
            Whether the folder being shared can be downloaded when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param allow_preview:
            Whether the folder being shared can be previewed when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :param vanity_name:
            Defines a custom vanity name to use in the shared link URL, eg. https://app.box.com/v/my-custom-vanity-name.
            If this parameter is None, the standard shared link URL will be used.
        :param kwargs:
            Used to fulfill the contract of overriden method
        :return:
            The updated object with shared link.
            Returns a new object of the same type, without modifying the original object passed as self.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the folder.
        """
        # pylint:disable=arguments-differ
        return super().create_shared_link(
            access=access,
            etag=etag,
            unshared_at=unshared_at,
            allow_download=allow_download,
            allow_preview=allow_preview,
            password=password,
            vanity_name=vanity_name
        )

    @api_call
    def get_shared_link(
            self,
            *,
            access: Optional[str] = None,
            etag: Optional[str] = None,
            unshared_at: Union[datetime, str, None] = SDK_VALUE_NOT_SET,
            allow_download: Optional[bool] = None,
            allow_preview: Optional[bool] = None,
            password: Optional[str] = None,
            vanity_name: Optional[str] = None,
            **kwargs: Any
    ) -> 'str':
        """
        Baseclass override.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :param etag:
            If specified, instruct the Box API to create the link only if the current version's etag matches.
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
            The time portion can be omitted, which defaults to midnight (00:00:00) on that date.
        :param allow_download:
            Whether the folder being shared can be downloaded when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param allow_preview:
            Whether the folder being shared can be previewed when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :param vanity_name:
            Defines a custom vanity name to use in the shared link URL, eg. https://app.box.com/v/my-custom-vanity-name.
            If this parameter is None, the standard shared link URL will be used.
        :param kwargs:
            Used to fulfill the contract of overriden method
        :returns:
            The URL of the shared link.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the folder.
        """
        # pylint:disable=arguments-differ
        return super().get_shared_link(
            access=access,
            etag=etag,
            unshared_at=unshared_at,
            allow_download=allow_download,
            allow_preview=allow_preview,
            password=password,
            vanity_name=vanity_name
        )

    @api_call
    def add_collaborator(
            self,
            collaborator: Union[User, Group, str],
            role: 'CollaborationRole',
            notify: bool = False,
            can_view_path: bool = False
    ) -> 'Collaboration':
        """Add a collaborator to the folder

        :param collaborator:
            collaborator to add. It may be a User, Group, or email address (unicode string)
        :param role:
            The collaboration role
        :param notify:
            Whether to send a notification email to the collaborator
        :param can_view_path:
            Whether view path collaboration feature is enabled or not. Note - only
            folder owners can create collaborations with can_view_path.
        :return:
            The new collaboration
        """
        collaborator_helper = _Collaborator(collaborator)
        url = self._session.get_url('collaborations')
        item = {'id': self._object_id, 'type': 'folder'}
        access_key, access_value = collaborator_helper.access
        accessible_by = {
            access_key: access_value,
            'type': collaborator_helper.type,
        }
        body_params = {
            'item': item,
            'accessible_by': accessible_by,
            'role': role,
        }
        if can_view_path:
            body_params['can_view_path'] = True
        data = json.dumps(body_params)
        params = {'notify': notify}
        box_response = self._session.post(url, expect_json_response=True, data=data, params=params)
        collaboration_response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=collaboration_response,
        )

    @api_call
    def create_web_link(
            self,
            target_url: str,
            name: Optional[str] = None,
            description: Optional[str] = None
    ) -> 'WebLink':
        """
        Create a WebLink with a given url.

        :param target_url:
            The url the web link points to.
        :param name:
            The name of the web link. Optional, the API will give it a default if not specified.
        :param description:
            Description of the web link
        :return:
            A :class:`WebLink` object.
        """
        url = self._session.get_url('web_links')
        web_link_attributes = {
            'url': target_url,
            'parent': {
                'id': self.object_id
            }
        }
        if name is not None:
            web_link_attributes['name'] = name
        if description is not None:
            web_link_attributes['description'] = description
        response = self._session.post(url, data=json.dumps(web_link_attributes)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response
        )

    @api_call
    def delete(
            self,
            *,
            recursive: bool = True,
            etag: Optional[str] = None,
            **kwargs
    ) -> bool:
        """Base class override. Delete the folder.

        :param recursive:
            Whether or not the folder should be deleted if it isn't empty.
        :param etag:
            If specified, instruct the Box API to delete the folder only if the current version's etag matches.
        :returns:
            Whether or not the update was successful.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the folder.
        """
        # pylint:disable=arguments-differ,arguments-renamed
        return super().delete(params={'recursive': recursive}, etag=etag, **kwargs)

    @api_call
    def get_metadata_cascade_policies(
            self,
            owner_enterprise: 'Enterprise' = None,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the metadata cascade policies current applied to the folder.

        :param owner_enterprise:
            Which enterprise's metadata templates to get cascade policies for.  This defauls to the current
            enterprise.
        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the cascade policies attached on the folder.
        """
        additional_params = {
            'folder_id': self.object_id,
        }
        if owner_enterprise is not None:
            additional_params['owner_enterprise_id'] = owner_enterprise.object_id

        return MarkerBasedObjectCollection(
            url=self._session.get_url('metadata_cascade_policies'),
            session=self._session,
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def cascade_metadata(self, metadata_template: 'MetadataTemplate') -> 'MetadataCascadePolicy':
        """
        Create a metadata cascade policy to apply the metadata instance values on the folder for the given metadata
        template to all files within the folder.

        :param metadata_template:
            The metadata template to cascade values for
        :returns:
            The created metadata cascade policy
        """
        url = self._session.get_url('metadata_cascade_policies')

        body = {
            'folder_id': self.object_id,
            'scope': metadata_template.scope,
            'templateKey': metadata_template.template_key,
        }

        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(self._session, response)

    @api_call
    def create_lock(self) -> 'FolderLock':
        """
        Creates a folder lock on a folder, preventing it from being moved and/or deleted.

        :returns:
            The created folder lock
        """
        url = self._session.get_url('folder_locks')

        body = {
            'folder': {
                'type': 'folder',
                'id': self.object_id
            },
            'locked_operations': {
                'move': True,
                'delete': True
            }
        }

        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(self._session, response)

    @api_call
    def get_locks(self) -> 'BoxObjectCollection':
        """
        Lists all folder locks for a given folder.

        :returns:
            The collection of locks for a folder.
        """
        url = self._session.get_url('folder_locks')

        additional_params = {
            'folder_id': self.object_id,
        }

        return MarkerBasedObjectCollection(
            url=url,
            session=self._session,
            additional_params=additional_params,
            return_full_pages=False,
        )



import json
import pytest

from boxsdk.config import API
from boxsdk.object.trash import Trash


@pytest.fixture()
def test_trash(mock_box_session):
    return Trash(mock_box_session)


@pytest.fixture(params=('file', 'folder', 'web_link'))
def test_item_and_response(
        test_file,
        mock_file_response,
        test_folder,
        mock_folder_response,
        test_web_link,
        mock_web_link_response,
        request,
):
    if request.param == 'file':
        return test_file, mock_file_response
    if request.param == 'web_link':
        return test_web_link, mock_web_link_response
    return test_folder, mock_folder_response


def test_get_from_trash(test_item_and_response, test_trash, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type + "s"}/{test_item.object_id}/trash'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'created_at': '2015-05-07T14:31:16-07:00',
        'modified_at': '2015-05-07T14:31:16-07:00',
        'created_by': {
            'type': 'user',
            'id': '11111',
        },
    }
    trashed_item_info = test_trash.get_item(
        item=test_item, fields=['created_at', 'modified_at'])
    mock_box_session.get.assert_called_once_with(
        expected_url, params={'fields': 'created_at,modified_at'})
    assert isinstance(trashed_item_info, test_item.__class__)
    assert trashed_item_info.object_type == test_item.object_type
    assert trashed_item_info.object_id == test_item.object_id
    assert trashed_item_info.created_at == '2015-05-07T14:31:16-07:00'
    assert trashed_item_info.modified_at == '2015-05-07T14:31:16-07:00'
    assert trashed_item_info.created_by['type'] == 'user'
    assert trashed_item_info.created_by['id'] == '11111'


def test_restore_from_trash(test_item_and_response, test_trash, test_folder, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    new_name = 'New Name'
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type + "s"}/{test_item.object_id}'
    value = json.dumps({
        'name': new_name,
        'parent': {
            'id': test_folder.object_id,
        },
    })
    mock_box_session.post.return_value.json.return_value = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'created_at': '2015-05-07T14:31:16-07:00',
        'modified_at': '2015-05-07T14:31:16-07:00',
        'created_by': {
            'type': 'user',
            'id': '11111',
        },
    }
    restored_item = test_trash.restore_item(test_item, new_name, test_folder, [
                                            'created_at', 'modified_at'])
    mock_box_session.post.assert_called_once_with(expected_url, data=value, params={
                                                  'fields': 'created_at,modified_at'})
    assert isinstance(restored_item, test_item.__class__)
    assert restored_item.object_type == test_item.object_type
    assert restored_item.object_id == test_item.object_id
    assert restored_item.created_at == '2015-05-07T14:31:16-07:00'
    assert restored_item.modified_at == '2015-05-07T14:31:16-07:00'
    assert restored_item.created_by['type'] == 'user'
    assert restored_item.created_by['id'] == '11111'


def test_permanently_delete(test_item_and_response, test_trash, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type + "s"}/{test_item.object_id}/trash'
    mock_box_session.delete.return_value.ok = True
    info = test_trash.permanently_delete_item(test_item)
    mock_box_session.delete.assert_called_once_with(
        expected_url, expect_json_response=False)
    assert info is True


def test_get_trashed_items(test_item_and_response, test_trash, mock_box_session):
    test_item, _ = test_item_and_response
    item_name = 'Test Trashed Item'
    expected_url = f'{API.BASE_API_URL}/folders/trash/items'
    mock_trash = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'name': 'Test Trashed Item'
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 5,
        'offset': 0,
        'limit': 100,
        'entries': [mock_trash]
    }
    trashed_items = test_trash.get_items(fields=['name'])
    trashed_item = trashed_items.next()
    mock_box_session.get.assert_called_once_with(
        expected_url, params={'fields': 'name', 'offset': None})
    assert isinstance(trashed_item, test_item.__class__)
    assert trashed_item.type == mock_trash['type']
    assert trashed_item.id == mock_trash['id']
    assert trashed_item.name == item_name


def test_get_trashed_items_with_sort(test_item_and_response, test_trash, mock_box_session):
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/folders/trash/items'
    mock_trash = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'name': 'Test Trashed Item'
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 5,
        'offset': 0,
        'limit': 100,
        'entries': [mock_trash]
    }
    trashed_items = test_trash.get_items(fields=['name'], sort='name', direction='ASC')
    trashed_item = trashed_items.next()
    mock_box_session.get.assert_called_once_with(
        expected_url, params={'direction': 'ASC', 'sort': 'name', 'offset': None, 'fields': 'name'})
    assert isinstance(trashed_item, test_item.__class__)


def test_get_trashed_items_with_marker(test_item_and_response, test_trash, mock_box_session):
    test_item, _ = test_item_and_response
    item_name = 'Test Trashed Item'
    expected_url = f'{API.BASE_API_URL}/folders/trash/items'
    mock_trash = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'name': 'Test Trashed Item'
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'next_marker': 2345,
        'entries': [mock_trash]
    }
    trashed_items = test_trash.get_items(
        fields=['name'], limit=100, marker=1234, use_marker=True)
    trashed_item = trashed_items.next()

    mock_box_session.get.assert_called_once_with(expected_url, params={
                                                 'limit': 100, 'usemarker': True, 'marker': 1234, 'fields': 'name'})
    assert isinstance(trashed_item, test_item.__class__)
    assert trashed_item.type == mock_trash['type']
    assert trashed_item.id == mock_trash['id']
    assert trashed_item.name == item_name



import uuid
import sqlalchemy
from sqlalchemy.orm import relationship
from test.functional.mock_box.db_model import DbModel


_user_group_table = sqlalchemy.Table(
    'user_group_association',
    DbModel.metadata,
    sqlalchemy.Column('group_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('box_group.id')),
    sqlalchemy.Column('user_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('box_user.id'))
)


class GroupModel(DbModel):
    """DB Model for Box groups."""
    __tablename__ = 'box_group'

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)  # pylint:disable=invalid-name
    group_id = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)
    name = sqlalchemy.Column(sqlalchemy.String(255))
    users = relationship('UserModel', secondary=_user_group_table, backref='groups')



def test_get(mock_collection, mock_box_session):
    expected_url = mock_collection.get_url()
    collection_name = 'Favorites'

    mock_box_session.get.return_value.json.return_value = {
        'type': 'collection',
        'id': mock_collection.object_id,
        'name': collection_name,
        'collection_type': 'favorites'
    }

    fetched_collection = mock_collection.get()

    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert fetched_collection.name == collection_name


def test_get_items(mock_collection, mock_box_session):
    expected_url = mock_collection.get_url('items')
    item_id1 = '12345'
    item_id2 = '56789'

    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'offset': 0,
        'total_count': 2,
        'entries': [
            {
                'type': 'folder',
                'id': item_id1
            },
            {
                'type': 'file',
                'id': item_id2
            }
        ]
    }

    items = mock_collection.get_items()

    item1 = items.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': 0})
    assert item1.type == 'folder'
    assert item1.object_id == item_id1

    item2 = items.next()
    assert item2.type == 'file'
    assert item2.object_id == item_id2



from datetime import datetime
import uuid
import sqlalchemy
from sqlalchemy.orm import backref, relationship
from test.functional.mock_box.db_model import DbModel


class FolderModel(DbModel):
    """DB Model for Box folders."""
    __tablename__ = 'box_folder'

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)  # pylint:disable=invalid-name
    folder_id = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: str(uuid.uuid4().int))
    name = sqlalchemy.Column(sqlalchemy.String(255))
    files = relationship('FileModel', backref='parent', cascade='save-update, delete')
    parent_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_folder.id'))
    folders = relationship('FolderModel', backref=backref('parent', remote_side=[id]), cascade='save-update, delete')
    created_by_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_user.id'))
    created_by = relationship('UserModel', foreign_keys=[created_by_id])
    created_at = sqlalchemy.Column(sqlalchemy.DateTime, default=datetime.now)
    modified_at = sqlalchemy.Column(sqlalchemy.DateTime, onupdate=datetime.now)
    owned_by_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_user.id'))
    collaborations = relationship('CollaborationModel', backref='item', cascade='save-update, delete')
    shared_link = relationship(
        'ShareFolderModel',
        backref=backref('shared_folder', remote_side=[id]),
        cascade='save-update, delete',
    )
    sync_state = sqlalchemy.Column(sqlalchemy.Enum('not_synced', 'partially_synced', 'synced'), default='not_synced')
    sequence_id = sqlalchemy.Column(sqlalchemy.Integer, nullable=False)
    etag = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)

    __mapper_args__ = {
        "version_id_col": sequence_id,
    }



import json
from typing import TYPE_CHECKING, Any, Iterable, Optional, Union, List

from .base_endpoint import BaseEndpoint
from .base_api_json_object import BaseAPIJSONObject
from ..util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.session.session import Session


class BaseObject(BaseEndpoint, BaseAPIJSONObject):
    """A Box API endpoint for interacting with a Box object."""

    def __init__(self, session: 'Session', object_id: str, response_object: dict = None):
        """
        :param session:
            The Box session used to make requests.
        :param object_id:
            The Box ID for the object.
        :param response_object:
            A JSON object representing the object returned from a Box API request.
        """
        super().__init__(session=session, response_object=response_object)
        self._object_id = object_id

    @property
    def _description(self) -> str:
        """
        Base class override.  Return a description for the object.
        """
        if 'name' in self._response_object:
            return f'{self._object_id} ({self.name})'  # pylint:disable=no-member
        return self._object_id

    def get_url(self, *args: Any) -> str:
        """
        Base class override.
        Return the given object's URL, appending any optional parts as specified by args.
        """
        return super().get_url(f'{self._item_type}s', self._object_id, *args)

    def get_type_url(self) -> str:
        """
        Return the URL for type of the given resource.
        """
        return super().get_url(f'{self._item_type}s')

    @property
    def object_id(self) -> str:
        """
        Return the Box ID for the object.
        """
        return self._object_id

    @api_call
    def get(self, *, fields: Iterable[str] = None, headers: dict = None, **_kwargs) -> Any:
        """
        Get information about the object, specified by fields. If fields is None, return the default fields.

        :param fields:
            List of fields to request.
        :param headers:
            Additional headers to send with the request.
        :return:
            An object of the same type that has the requested information.
        """
        url = self.get_url()
        params = {'fields': ','.join(fields)} if fields else None
        box_response = self._session.get(url, params=params, headers=headers)
        return self.translator.translate(
            session=self._session,
            response_object=box_response.json(),
        )

    @api_call
    def update_info(
            self,
            *,
            data: Union[dict, List[dict]],
            params: Optional[dict] = None,
            headers: Optional[dict] = None,
            **kwargs: Any
    ) -> Any:
        """Update information about this object.

        Send a PUT to the object's base endpoint to modify the provided
        attributes.

        :param data:
            The updated information about this object.
            Must be JSON serializable.
            Update the object attributes in data.keys(). The semantics of the
            values depends on the type and attributes of the object being
            updated. For details on particular semantics, refer to the Box
            developer API documentation <https://developer.box.com/>.
        :param params:
            (optional) Query string parameters for the request.
        :param headers:
            (optional) Extra HTTP headers for the request.
        :param kwargs:
            Optional arguments that ``put`` takes.
        :return:
            The updated object.
            Return a new object of the same type, without modifying the
            original object passed as self.
            Construct the new object with all the default attributes that are
            returned from the endpoint.
        """
        url = self.get_url()
        box_response = self._session.put(url, data=json.dumps(data), params=params, headers=headers, **kwargs)
        if 'expect_json_response' in kwargs and not kwargs['expect_json_response']:
            return box_response.ok

        return self.translator.translate(
            session=self._session,
            response_object=box_response.json(),
        )

    @api_call
    def delete(self, *, params: Optional[dict] = None, headers: Optional[dict] = None, **_kwargs) -> bool:
        """ Delete the object.

        :param params:
            Additional parameters to send with the request. Can be None
        :param headers:
            Any customer headers to send with the request. Can be None
        :returns:
            Whether or not the delete was successful.
        :raises:
            :class:`BoxAPIException` in case of unexpected errors.
        """
        url = self.get_url()

        box_response = self._session.delete(url, expect_json_response=False, params=params or {}, headers=headers)
        return box_response.ok

    def __eq__(self, other: Any) -> bool:
        """Equality as determined by object id and type"""
        if isinstance(other, BaseObject):
            # Two objects are considered the same if they have the same address in the API
            return self.get_url() == other.get_url()

        return NotImplemented

    def __ne__(self, other: Any) -> bool:
        """Equality as determined by object id and type"""
        return not self == other

    def __hash__(self) -> int:
        return hash((self._object_id, self._item_type))

    def clone(self, session: 'Session' = None) -> 'BaseObject':
        """Base class override."""
        return self.__class__(
            session or self._session,
            self._object_id,
            self._response_object,
        )



from .dict_page import DictPage
from .marker_based_object_collection import MarkerBasedObjectCollection


class MarkerBasedDictCollection(MarkerBasedObjectCollection):
    """Represents a marker-based collection of simple dicts, which are not translated into objects."""
    _page_constructor = DictPage



from unittest.mock import ANY
import json
import pytest

from boxsdk.config import API
from boxsdk.object.file import File
from boxsdk.object.user import User
from boxsdk.object.search import MetadataSearchFilters, MetadataSearchFilter, SearchScope, TrashContent


@pytest.fixture
def search_query():
    return 'myquery'


@pytest.fixture(params=(1, 20, 100))
def search_limit(request):
    return request.param


@pytest.fixture(params=(0, 10))
def search_offset(request):
    return request.param


@pytest.fixture(params=(None, 'file', 'folder'))
def search_result_type(request):
    return request.param


@pytest.fixture(params=(None, ('name',), ('name', 'description')))
def search_content_types(request):
    return request.param


@pytest.fixture
def search_value_based_filters():
    metadata_filters = MetadataSearchFilters()
    metadata_filter = MetadataSearchFilter(template_key='mytemplate', scope='enterprise')
    metadata_filter.add_value_based_filter(field_key='myfield', value='myvalue')
    metadata_filters.add_filter(metadata_filter)
    return metadata_filters


@pytest.fixture(params=(
    {'gt_value': 'mygtvalue'},
    {'lt_value': 'myltvalue'},
    {'gt_value': 'mygtvalue', 'lt_value': 'myltvalue'}
))
def search_range_filters(request):
    metadata_filters = MetadataSearchFilters()
    metadata_filter = MetadataSearchFilter(template_key='mytemplate', scope='enterprise')
    filter_params = {'field_key': 'myfield'}
    filter_params.update(request.param)
    metadata_filter.add_range_filter(**filter_params)
    metadata_filters.add_filter(metadata_filter)
    return metadata_filters


@pytest.fixture
def search_entries():
    return [
        {'id': '1234', 'type': 'file'}
    ]


@pytest.fixture
def search_response(search_entries):
    return {
        'entries': search_entries,
        'total_count': len(search_entries),
        'limit': 20,
        'offset': 0
    }


@pytest.fixture
def metadata_query_response():
    return {
        'entries': [
            {
                'type': 'file',
                'id': '1244738582',
                'name': 'Very Important.docx',
                'metadata': {
                    'enterprise_67890': {
                        'catalogImages': {
                            '$parent': 'file_50347290',
                            '$version': 2,
                            '$template': 'catalogImages',
                            '$scope': 'enterprise_67890',
                            'photographer': 'Bob Dylan'
                        }
                    }
                }
            },
            {
                'type': 'folder',
                'id': '124242482',
                'name': 'Also Important.docx',
                'metadata': {
                    'enterprise_67890': {
                        'catalogImages': {
                            '$parent': 'file_50427291',
                            '$version': 2,
                            '$template': 'catalogImages',
                            '$scope': 'enterprise_67890',
                            'photographer': 'Bob Dylan'
                        }
                    }
                }
            }
        ],
        'limit': 2,
        'next_marker': ''
    }


@pytest.fixture
def search_with_shared_links_entries():
    return [
        {
            'accessible_via_shared_link': 'https://www.box.com/s/vspke7y05sb214wjokpk',
            'item': {'id': '1234', 'type': 'file'},
            'type': 'search_result'
        },
        {
            'accessible_via_shared_link': None,
            'item': {'id': '1234', 'type': 'file'},
            'type': 'search_result'
        }
    ]


@pytest.fixture
def search_with_shared_links_response(search_with_shared_links_entries):
    entries = search_with_shared_links_entries
    return {
        'entries': entries,
        'total_count': len(entries),
        'limit': 20,
        'offset': 0
    }


class Matcher:
    def __init__(self, compare, some_obj):
        self.compare = compare
        self.some_obj = some_obj

    def __eq__(self, other):
        return self.compare(self.some_obj, other)


def compare_params(self, other):
    if not isinstance(self, dict) or not isinstance(other, dict):
        return False
    for key in self:
        # We need to ensure that the JSON-encoded mdfilters matches regardless of key order
        if key == 'mdfilters':
            if json.loads(self['mdfilters']) != json.loads(other['mdfilters']):
                return False
        # For other keys, just ensure that they are equal
        elif key in ('type', 'content_types'):
            return self[key] is None or self[key] == other[key]
        else:
            if self[key] != other[key]:
                return False
    return True


def test_query_with_value_based_filters(
        mock_box_session,
        make_mock_box_request,
        test_search,
        search_query,
        search_limit,
        search_offset,
        search_value_based_filters,
        search_response,
        search_entries,
        search_result_type,
        search_content_types,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value, _ = make_mock_box_request(response=search_response)
    response = test_search.query(
        search_query,
        limit=search_limit,
        offset=search_offset,
        metadata_filters=search_value_based_filters,
        result_type=search_result_type,
        content_types=search_content_types,
    )
    for actual, expected in zip(response, [File(mock_box_session, item['id'], item) for item in search_entries]):
        assert actual == expected

    mock_box_session.get.assert_called_once_with(
        test_search.get_url(),
        params=Matcher(compare_params, {
            'query': search_query,
            'limit': search_limit,
            'mdfilters': json.dumps(search_value_based_filters.as_list()),
            'offset': search_offset,
            'type': search_result_type,
            'content_types': ','.join(search_content_types) if search_content_types else search_content_types,
        })
    )


def test_query_with_range_filters(
        mock_box_session,
        make_mock_box_request,
        test_search,
        search_query,
        search_limit,
        search_offset,
        search_range_filters,
        search_response,
        search_entries,
        search_result_type,
        search_content_types,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value, _ = make_mock_box_request(response=search_response)
    response = test_search.query(
        search_query,
        limit=search_limit,
        offset=search_offset,
        metadata_filters=search_range_filters,
        result_type=search_result_type,
        content_types=search_content_types,
    )
    for actual, expected in zip(response, [File(mock_box_session, item['id'], item) for item in search_entries]):
        assert actual == expected

    mock_box_session.get.assert_called_once_with(
        test_search.get_url(),
        params=Matcher(compare_params, {
            'query': search_query,
            'limit': search_limit,
            'mdfilters': json.dumps(search_range_filters.as_list()),
            'offset': search_offset,
            'type': search_result_type,
            'content_types': ','.join(search_content_types) if search_content_types else search_content_types,
        })
    )


@pytest.mark.parametrize('kwargs, params', [
    ({'scope': SearchScope.ENTERPRISE}, {'scope': 'enterprise_content'}),
    ({'scope': SearchScope.USER}, {'scope': 'user_content'}),
    ({'created_at_range': (None, '2018-01-01T00:00:00Z')}, {'created_at_range': ',2018-01-01T00:00:00Z'}),
    ({'created_at_range': ('2015-02-03T12:00:00-08:00', None)}, {'created_at_range': '2015-02-03T12:00:00-08:00,'}),
    ({'created_at_range': ('2012-01-01T00:00:00Z', '2012-12-31T11:59:59Z')}, {'created_at_range': '2012-01-01T00:00:00Z,2012-12-31T11:59:59Z'}),
    ({'updated_at_range': (None, '2018-01-01T00:00:00Z')}, {'updated_at_range': ',2018-01-01T00:00:00Z'}),
    ({'updated_at_range': ('2015-02-03T12:00:00-08:00', None)}, {'updated_at_range': '2015-02-03T12:00:00-08:00,'}),
    ({'updated_at_range': ('2012-01-01T00:00:00Z', '2012-12-31T11:59:59Z')}, {'updated_at_range': '2012-01-01T00:00:00Z,2012-12-31T11:59:59Z'}),
    ({'size_range': (None, 123)}, {'size_range': ',123'}),
    ({'size_range': (123, None)}, {'size_range': '123,'}),
    ({'size_range': (123, 456)}, {'size_range': '123,456'}),
    ({'trash_content': TrashContent.NONE}, {'trash_content': 'non_trashed_only'}),
    ({'trash_content': TrashContent.ONLY}, {'trash_content': 'trashed_only'}),
    ({'sort': 'modified_at'}, {'sort': 'modified_at'}),
    ({'direction': 'DESC'}, {'direction': 'DESC'}),
])
def test_query_with_optional_parameters(
        mock_box_session,
        test_search,
        make_mock_box_request,
        search_query,
        search_response,
        search_entries,
        kwargs,
        params
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value, _ = make_mock_box_request(response=search_response)
    response = test_search.query(
        search_query,
        **kwargs
    )
    for actual, expected in zip(response, [File(mock_box_session, item['id'], item) for item in search_entries]):
        assert actual == expected

    expected_params = {
        'query': search_query,
    }
    expected_params.update(params)

    mock_box_session.get.assert_called_once_with(
        test_search.get_url(),
        params=Matcher(compare_params, expected_params)
    )


def test_query_with_owner_users(
        mock_box_session,
        test_search,
        make_mock_box_request,
        search_query,
        search_response,
        search_entries,
):
    # pylint:disable=redefined-outer-name
    user1 = User(mock_box_session, '987')
    user2 = User(mock_box_session, '654')
    mock_box_session.get.return_value, _ = make_mock_box_request(response=search_response)
    response = test_search.query(
        search_query,
        owner_users=[user1, user2],
    )
    for actual, expected in zip(response, [File(mock_box_session, item['id'], item) for item in search_entries]):
        assert actual == expected

    mock_box_session.get.assert_called_once_with(
        test_search.get_url(),
        params=Matcher(compare_params, {
            'query': search_query,
            'owner_user_ids': '987,654'
        })
    )


def test_metadata_query(
        mock_box_session,
        make_mock_box_request,
        test_search,
        metadata_query_response
):
    # pylint:disable=redefined-outer-name
    expected_url = f'{API.BASE_API_URL}/metadata_queries/execute_read'
    from_template = 'enterprise_12345.someTemplate'
    ancestor_folder_id = '5555'
    query = 'amount >= :arg'
    query_params = {'arg': 100}
    order_by = [
        {
            'field_key': 'amount',
            'direction': 'asc'
        }
    ]
    fields = ['type', 'id', 'name', 'metadata.enterprise_67890.catalogImages.$parent']
    limit = 2
    marker = 'AAAAAmVYB1FWec8GH6yWu2nwmanfMh07IyYInaa7DZDYjgO1H4KoLW29vPlLY173OKs'
    expected_data = {
        'limit': limit,
        'from': from_template,
        'ancestor_folder_id': ancestor_folder_id,
        'query': query,
        'query_params': query_params,
        'order_by': order_by,
        'marker': marker,
        'fields': fields
    }
    expected_headers = {b'Content-Type': b'application/json'}
    mock_box_session.post.return_value, _ = make_mock_box_request(response=metadata_query_response)
    items = test_search.metadata_query(
        from_template=from_template,
        ancestor_folder_id=ancestor_folder_id,
        query=query,
        query_params=query_params,
        order_by=order_by,
        marker=marker,
        limit=limit,
        fields=fields
    )
    item1 = items.next()
    item2 = items.next()
    mock_box_session.post.assert_called_once_with(expected_url, data=ANY, headers=expected_headers)
    assert dict(json.loads(mock_box_session.post.call_args[1]['data'])) == expected_data
    assert mock_box_session.post.call_args[1]['headers'] == expected_headers
    assert item1['type'] == 'file'
    assert item1['metadata']['enterprise_67890']['catalogImages']['$parent'] == 'file_50347290'
    assert item2['type'] == 'folder'
    assert item2['metadata']['enterprise_67890']['catalogImages']['$parent'] == 'file_50427291'


def test_range_filter_without_gt_and_lt_will_fail_validation():
    metadata_filter = MetadataSearchFilter(template_key='mytemplate', scope='enterprise')
    with pytest.raises(ValueError):
        metadata_filter.add_range_filter(field_key='mykey')


def test_start_search_filters(test_search):
    filters = test_search.start_metadata_filters()
    assert isinstance(filters, MetadataSearchFilters)


def test_make_single_metadata_filter(test_search):
    template_key = 'mytemplate'
    scope = 'myscope'
    metadata_filter = test_search.make_single_metadata_filter(template_key, scope)
    assert isinstance(metadata_filter, MetadataSearchFilter)
    filter_as_dict = metadata_filter.as_dict()
    assert filter_as_dict['templateKey'] == template_key
    assert filter_as_dict['scope'] == scope


def test_query_with_shared_links(
        mock_box_session,
        make_mock_box_request,
        search_content_types,
        search_limit,
        search_offset,
        search_query,
        search_result_type,
        search_value_based_filters,
        search_with_shared_links_entries,
        search_with_shared_links_response,
        test_search,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value, _ = make_mock_box_request(response=search_with_shared_links_response)
    response = test_search.query_with_shared_links(
        search_query,
        limit=search_limit,
        offset=search_offset,
        metadata_filters=search_value_based_filters,
        result_type=search_result_type,
        content_types=search_content_types,
    )

    for actual, expected in zip(response, [File(mock_box_session, entry['item']['id'], entry['item']) for entry in search_with_shared_links_entries]):
        assert actual.item == expected

    mock_box_session.get.assert_called_once_with(
        test_search.get_url(),
        params=Matcher(compare_params, {
            'query': search_query,
            'include_recent_shared_links': True,
            'limit': search_limit,
            'mdfilters': json.dumps(search_value_based_filters.as_list()),
            'offset': search_offset,
            'type': search_result_type,
            'content_types': ','.join(search_content_types) if search_content_types else search_content_types,
        })
    )
